/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Errors.h"
using pecunia::Error;
using std::exception;
using std::make_unique;
using std::string;
using std::vector;

#include <utility>
using std::move;

#include <internal/BackTrace.h>
using pecunia::internal::captureBackTrace;
using pecunia::internal::generateBackTrace;
#include <internal/TraceCapture.h>
using pecunia::internal::TraceCapture;


Error::Error() noexcept :
	Error{"pecunia::Error"}
{}

Error::Error(string message) noexcept : // cppcheck-suppress passedByValue
	exception{},
	message_{move(message)},
	trace_{make_unique<TraceCapture>(captureBackTrace())}
{}

Error::Error(const string& message, const exception& original) noexcept :
	Error{message + ' ' + original.what()}
{}

Error::Error(const Error& other) noexcept :
	message_{other.message_},
	trace_{make_unique<TraceCapture>(*other.trace_)}
{}

Error::Error(Error&& other) noexcept :
	message_{move(other.message_)},
	trace_{move(other.trace_)}
{}

Error::~Error() noexcept = default;

Error& Error::operator=(const Error& other) noexcept
{
	if (this != &other)
	{
		this->message_ = other.message_;
		this->trace_ = make_unique<TraceCapture>(*other.trace_);
	}
	return *this;
}

Error& Error::operator=(Error&& other) noexcept
{
	if (this != &other)
	{
		this->message_ = move(other.message_);
		this->trace_ = move(other.trace_);
	}
	return *this;
}

const char* Error::what() const noexcept
{
	return this->message_.c_str();
}

vector<string> Error::frameStack() const noexcept
{
	return generateBackTrace(*this->trace_);
}
