/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_ERRORS_H
#define PECUNIA_ERRORS_H

#include <exception>
#include <memory>
#include <string>
#include <vector>

#include <pecunia/pecunia_export.h>


namespace pecunia::internal
{

class TraceCapture;

}
namespace pecunia
{

#ifdef _WIN32
#	pragma warning(push)
// Justification: The warning "needs to have dll - interface to be used by clients of class" does
// not apply to inheirtence from STL types.
#	pragma warning(disable : 4251)
// Justification: The warning "needs to have dll - interface to be used by clients of class" does
// not apply to inheirtence from STL types.
#	pragma warning(disable : 4275)
#endif

/**
 * @brief A representation of a generalised software issue, complete with the location of the
 *  problem.
 */
class PECUNIA_EXPORT Error : public std::exception
{
	std::string message_;
	std::unique_ptr<internal::TraceCapture> trace_;

public:
	Error() noexcept;
	/**
	 * @brief Message initialisation constructor.
	 *
	 * @param message The message describing the cause of the error.
	 */
	explicit Error(std::string message) noexcept;
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @param message The specific message describing the cause of the error.
	 * @param original The original error that caused the error.
	 */
	Error(const std::string& message, const std::exception& original) noexcept;
	Error(const Error& other) noexcept;
	Error(Error&& other) noexcept;
	~Error() noexcept override;
	Error& operator=(const Error& other) noexcept;
	Error& operator=(Error&& other) noexcept;
	/**
	 * @brief An accessor to the message describing the cause of the error.
	 */
	[[nodiscard]]
	const char* what() const noexcept override;
	/**
	 * @brief Provides the frame-stack leading to the source of the error.
	 *
	 * @return A container of each line leading to the source of the error. The specific formatting
	 * is dependent upon the compiler.
	 */
	std::vector<std::string> frameStack() const noexcept;
};
#ifdef _WIN32
#	pragma warning(pop)
#endif

/**
 * @brief Represents a flaw in the logic or a precondition of a function.
 */
struct PECUNIA_EXPORT LogicError : Error
{
	using Error::Error;
};
/**
 * @brief Represents an argument that is invalid for the state of the function.
 */
struct PECUNIA_EXPORT InvalidArgument : LogicError
{
	using LogicError::LogicError;
};
/**
 * @brief Represents an argument that outside the realm of the purpose of a function.
 */
struct PECUNIA_EXPORT DomainError : LogicError
{
	using LogicError::LogicError;
};
/**
 * @brief Represents an argument that is out of range for a container.
 */
struct PECUNIA_EXPORT OutOfRange : LogicError
{
	using LogicError::LogicError;
};
/**
 * @brief Represents a flaw that prevents a function from fulfilling its postcondition or purpose.
 */
struct PECUNIA_EXPORT RuntimeError : Error
{
	using Error::Error;
};
/**
 * @brief Represents a flaw that would trigger an overflow operation.
 */
struct PECUNIA_EXPORT OverflowError : RuntimeError
{
	using RuntimeError::RuntimeError;
};
/**
 * @brief Represents a flaw that would trigger an underflow operation.
 */
struct PECUNIA_EXPORT UnderflowError : RuntimeError
{
	using RuntimeError::RuntimeError;
};

}

#endif
