/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "FloatingPoint.h"
using namespace pecunia::literals;
using std::string;
using std::to_string;
using pecunia::FloatingPoint;
using pecunia::FloatingPointBase;
using pecunia::RealNumber;

#include <cmath>
using std::abs;
using std::ceil;
using std::fabs;
using std::floor;
using std::fmod;
using std::modf;
using std::pow;
using std::sqrt;
#include <cfenv>
using std::feclearexcept;
using std::fetestexcept;
#include <exception>
using std::exception;
#include <utility>
using std::move;

#include "Errors.h"
using pecunia::DomainError;
using pecunia::OverflowError;
using pecunia::UnderflowError;


namespace
{

FloatingPoint testExceptions(const FloatingPointBase& value)
{
	int exceptions{fetestexcept(FE_ALL_EXCEPT)};

	if (exceptions & FE_DIVBYZERO)
		throw DomainError{"The floating-point operation resulted in a divide by zero."};

	if (exceptions & FE_UNDERFLOW)
		throw UnderflowError{
			"The floating-point operation resulted in a value that is too small."
		};

	if (exceptions & FE_OVERFLOW)
		throw OverflowError{
			"The floating-point operation resulted in a value that is too large."
		};

	if (exceptions & FE_INVALID)
		throw DomainError{
			"The floating-point operation resulted in a value that is not valid."
		};
	return FloatingPoint{value};
}

}

FloatingPoint::FloatingPoint() :
	FloatingPoint{static_cast<FloatingPointBase>(0)}
{}

FloatingPoint::FloatingPoint(const FloatingPointBase& value) :
	value_{value}
{}

FloatingPoint::FloatingPoint(const int32_t& value) :
	FloatingPoint{static_cast<FloatingPointBase>(value)}
{}

FloatingPoint::FloatingPoint(const int64_t& value) :
	FloatingPoint{static_cast<FloatingPointBase>(value)}
{}

FloatingPoint FloatingPoint::operator+(const FloatingPoint& other) const
{
	auto value{*this};
	return value += other;
}

FloatingPoint FloatingPoint::operator-(const FloatingPoint& other) const
{
	auto value{*this};
	return value -= other;
}

FloatingPoint FloatingPoint::operator*(const FloatingPoint& other) const
{
	auto value{*this};
	return value *= other;
}

FloatingPoint FloatingPoint::operator/(const FloatingPoint& divisor) const
{
	auto dividend{*this};
	return dividend /= divisor;
}

FloatingPoint& FloatingPoint::operator+=(const FloatingPoint& other)
{
	feclearexcept(FE_ALL_EXCEPT);
	const auto copy{this->value_};

	try
	{
		testExceptions(this->value_ += other.value_);
	}
	catch (const exception&)
	{
		this->value_ = copy;
		throw;
	}
	return *this;
}

FloatingPoint& FloatingPoint::operator-=(const FloatingPoint& other)
{
	feclearexcept(FE_ALL_EXCEPT);
	const auto copy{this->value_};

	try
	{
		testExceptions(this->value_ -= other.value_);
	}
	catch (const exception&)
	{
		this->value_ = copy;
		throw;
	}
	return *this;
}

FloatingPoint& FloatingPoint::operator*=(const FloatingPoint& other)
{
	feclearexcept(FE_ALL_EXCEPT);
	const auto copy{this->value_};

	try
	{
		testExceptions(this->value_ *= other.value_);
	}
	catch (const exception&)
	{
		this->value_ = copy;
		throw;
	}
	return *this;
}

FloatingPoint& FloatingPoint::operator/=(const FloatingPoint& divisor)
{
	feclearexcept(FE_ALL_EXCEPT);
	const auto copy{this->value_};

	try
	{
		testExceptions(this->value_ /= divisor.value_);
	}
	catch (const exception&)
	{
		this->value_ = copy;
		throw;
	}
	return *this;
}

bool FloatingPoint::operator<(const FloatingPoint& other) const noexcept
{
	return this->value_ < other.value_;
}

bool FloatingPoint::operator>(const FloatingPoint& other) const noexcept
{
	return this->value_ > other.value_;
}

bool FloatingPoint::operator<(const FloatingPointBase& other) const noexcept
{
	return this->value_ < other;
}

bool FloatingPoint::operator>(const FloatingPointBase& other) const noexcept
{
	return this->value_ > other;
}

FloatingPoint FloatingPoint::operator-() const
{
	feclearexcept(FE_ALL_EXCEPT);
	return testExceptions(-this->value_);
}

const FloatingPointBase& FloatingPoint::underlying() const noexcept
{
	return this->value_;
}

RealNumber FloatingPoint::decompose(const uint8_t digits) const
{
	feclearexcept(FE_ALL_EXCEPT);
	FloatingPointBase whole;
	const auto fraction{testExceptions(modf(this->value_, &whole))};
	const auto shiftPrecision{10.0_fp .power(digits)};
	const auto shiftedFractionWhole{fraction * shiftPrecision};
	const auto shiftedFraction{shiftedFractionWhole / shiftPrecision};
	return RealNumber{whole, shiftedFraction};
}

FloatingPoint FloatingPoint::floor() const
{
	feclearexcept(FE_ALL_EXCEPT);
	return testExceptions(::floor(this->value_));
}

FloatingPoint FloatingPoint::ceiling() const
{
	feclearexcept(FE_ALL_EXCEPT);
	return testExceptions(ceil(this->value_));
}

FloatingPoint FloatingPoint::absoluteValue() const
{
	feclearexcept(FE_ALL_EXCEPT);
	return testExceptions(abs(this->value_));
}

FloatingPoint FloatingPoint::modulo(const FloatingPoint& divisor) const
{
	feclearexcept(FE_ALL_EXCEPT);
	return testExceptions(fmod(this->value_, divisor.value_));
}

FloatingPoint FloatingPoint::power(const int64_t& exponent) const
{
	feclearexcept(FE_ALL_EXCEPT);
	return testExceptions(pow(this->value_, exponent));
}

FloatingPoint FloatingPoint::power(const FloatingPoint& exponent) const
{
	feclearexcept(FE_ALL_EXCEPT);
	return testExceptions(pow(this->value_, exponent.value_));
}

bool FloatingPoint::isEqual(const FloatingPoint& other, const uint8_t digits) const
{
	const auto difference{(*this - other).absoluteValue()};
	const auto shiftPrecision{10.0_fp .power(digits)};
	const auto shiftedDifference{difference * shiftPrecision};
	const auto differenceInteger{static_cast<int64_t>(shiftedDifference.underlying())};
	const auto isEqualEnough{differenceInteger == 0};
	return isEqualEnough;
}

FloatingPoint FloatingPoint::reciprocal() const
{
	return 1.0_fp / *this;
}

string FloatingPoint::toString() const
{
	return to_string(this->value_);
}

FloatingPoint FloatingPoint::squareRoot() const
{
	feclearexcept(FE_ALL_EXCEPT);
	return testExceptions(sqrt(this->value_));
}

bool pecunia::operator<(const FloatingPointBase lhs, const FloatingPoint& rhs)
{
	return FloatingPoint{lhs} < rhs;
}

bool pecunia::operator>(const FloatingPointBase lhs, const FloatingPoint& rhs)
{
	return FloatingPoint{lhs} > rhs;
}

FloatingPoint pecunia::operator*(
	const FloatingPointBase multiplicand,
	const FloatingPoint& multiplier
)
{
	return FloatingPoint{multiplicand} * multiplier;
}

FloatingPoint pecunia::operator/(const FloatingPointBase dividend, const FloatingPoint& divisor)
{
	return FloatingPoint{dividend} / divisor;
}

FloatingPoint pecunia::literals::operator""_fp(long double value)
{
	return FloatingPoint{static_cast<FloatingPointBase>(value)};
}

