/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_FLOATINGPOINT_H
#define PECUNIA_FLOATINGPOINT_H

#include <string>

#include <pecunia/Sizes.hpp>
#include <pecunia/pecunia_export.h>


namespace pecunia
{

struct RealNumber;

/**
 * @brief Represents a safe type for working with floating-point operations.
 */
class PECUNIA_EXPORT FloatingPoint
{
	FloatingPointBase value_;

public:
	/**
	 * @brief Fully initialises the value to zero.
	 */
	FloatingPoint();
	/**
	 * @brief Fully initialises the value.
	 *
	 * @param value The value to set the floating-point to.
	 */
	FloatingPoint(const FloatingPointBase& value); // cppcheck-suppress noExplicitConstructor
	/**
	 * @brief Fully initialises the value.
	 *
	 * @param value The value to set the floating-point to.
	 */
	FloatingPoint(const std::int32_t& value); // cppcheck-suppress noExplicitConstructor
	/**
	 * @brief Fully initialises the value.
	 *
	 * @param value The value to set the floating-point to.
	 */
	FloatingPoint(const std::int64_t& value); // cppcheck-suppress noExplicitConstructor
	FloatingPoint(const FloatingPoint&) = default;
	FloatingPoint(FloatingPoint&&) noexcept = default;
	[[nodiscard]]
	FloatingPoint& operator=(const FloatingPoint&) = default;
	[[nodiscard]]
	FloatingPoint& operator=(FloatingPoint&&) noexcept = default;
	/**
	 * @brief Calculates the addition operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 *
	 * @param other The value to add to the current stored value.
	 *
	 * @return The result of the operation.
	 */
	[[nodiscard]]
	FloatingPoint operator+(const FloatingPoint& other) const;
	/**
	 * @brief Calculates the subtraction operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 *
	 * @param other The value to subtract from the current stored value.
	 *
	 * @return The result of the operation.
	 */
	[[nodiscard]]
	FloatingPoint operator-(const FloatingPoint& other) const;
	/**
	 * @brief Calculates the multiplication operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 *
	 * @param other The value to multiply the current stored value with.
	 *
	 * @return The result of the operation.
	 */
	[[nodiscard]]
	FloatingPoint operator*(const FloatingPoint& other) const;
	/**
	 * @brief Calculates the division operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @param divisor The value to divide the current stored value (dividend) with.
	 *
	 * @return The result of the operation.
	 */
	[[nodiscard]]
	FloatingPoint operator/(const FloatingPoint& divisor) const;
	/**
	 * @brief Calculates the addition and assignment operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 *
	 * @param other The value to add to the current stored value.
	 *
	 * @post The result of the operation is stored in the current object.
	 *
	 * @return The current object.
	 */
	FloatingPoint& operator+=(const FloatingPoint& other);
	/**
	 * @brief Calculates the subtraction and assignment operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 *
	 * @param other The value to subtract to the current stored value.
	 *
	 * @post The result of the operation is stored in the current object.
	 *
	 * @return The current object.
	 */
	FloatingPoint& operator-=(const FloatingPoint& other);
	/**
	 * @brief Calculates the multiplication and assignment operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 *
	 * @param other The value to multiply the current stored value with.
	 *
	 * @post The result of the operation is stored in the current object.
	 *
	 * @return The current object.
	 */
	FloatingPoint& operator*=(const FloatingPoint& other);
	/**
	 * @brief Calculates the division and assignment operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @param divisor The value to divide the current stored value (dividend) with.
	 *
	 * @post The result of the operation is stored in the current object.
	 *
	 * @return The current object.
	 */
	FloatingPoint& operator/=(const FloatingPoint& divisor);
	bool operator==(const FloatingPoint& other) const = delete;
	bool operator==(const FloatingPointBase& other) const = delete;
	/**
	 * @brief Relationship operator checking the less-than status.
	 *
	 * @param other The value to check the relationship with the current value.
	 *
	 * @return Gives true when the current value is less-than the supplied value.
	 */
	[[nodiscard]]
	bool operator<(const FloatingPoint& other) const noexcept;
	bool operator<=(const FloatingPoint& other) const = delete;
	/**
	 * @brief Relationship operator checking the greater-than status.
	 *
	 * @param other The value to check the relationship with the current value.
	 *
	 * @return Gives true when the current value is greater-than the supplied value.
	 */
	[[nodiscard]]
	bool operator>(const FloatingPoint& other) const noexcept;
	bool operator>=(const FloatingPoint& other) const = delete;
	/**
	 * @brief Relationship operator checking the less-than status.
	 *
	 * @param other The value to check the relationship with the current value.
	 *
	 * @return Gives true when the current value is less-than the supplied value.
	 */
	[[nodiscard]]
	bool operator<(const FloatingPointBase& other) const noexcept;
	bool operator<=(const FloatingPointBase& other) const = delete;
	/**
	 * @brief Relationship operator checking the greater-than status.
	 *
	 * @param other The value to check the relationship with the current value.
	 *
	 * @return Gives true when the current value is greater-than the supplied value.
	 */
	[[nodiscard]]
	bool operator>(const FloatingPointBase& other) const noexcept;
	bool operator>=(const FloatingPointBase& other) const = delete;
	/**
	 * @brief Negation operator.
	 *
	 * @return Gives the negation result of the current value.
	 */
	[[nodiscard]]
	FloatingPoint operator-() const;
	/**
	 * @brief The unsafe type represented by the floating-point value.
	 */
	[[nodiscard]]
	const FloatingPointBase& underlying() const noexcept;
	/**
	 * @brief Breaks a floating-point number into its whole and fractional components.
	 *
	 * @param digits The number of digits of precision to have in the fractional portion.
	 *
	 * @return The split floating-point number.
	 */
	[[nodiscard]]
	RealNumber decompose(const std::uint8_t digits = defaultRoundingDigits) const;
	/**
	 * @brief Performs the mathematical floor operation.
	 *
	 * @return The result of the floor operation on the current value.
	 */
	[[nodiscard]]
	FloatingPoint floor() const;
	/**
	 * @brief Performs the mathematical ceiling operation.
	 *
	 * @return The result of the ceiling operation on the current value.
	 */
	[[nodiscard]]
	FloatingPoint ceiling() const;
	/**
	 * @brief Performs the mathematical absolute value operation.
	 *
	 * @return The result of the absolute value operation on the current value.
	 */
	[[nodiscard]]
	FloatingPoint absoluteValue() const;
	/**
	 * @brief Calculates the modulo operation, i.e. division remainder.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @param divisor The value to perform the modulo of the current stored value (dividend) with.
	 *
	 * @return The result of the operation.
	 */
	[[nodiscard]]
	FloatingPoint modulo(const FloatingPoint& divisor) const;
	/**
	 * @brief Performs the exponentiation operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @param exponent The power to raise the base value to.
	 *
	 * @return The result of raising the current value, the base, to the power of the supplied
	 * exponent.
	 */
	[[nodiscard]]
	FloatingPoint power(const std::int64_t& exponent) const;
	/**
	 * @brief Performs the exponentiation operation.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @param exponent The power to raise the base value to.
	 *
	 * @return The result of raising the current value, the base, to the power of the supplied
	 * exponent.
	 */
	[[nodiscard]]
	FloatingPoint power(const FloatingPoint& exponent) const;
	/**
	 * @brief Determines if the supplied number is equal to within a specific number of digits.
	 *
	 * @param other The other floating-point value to compare the current instance to.
	 * @param digits The number of digits of precision to compare the two numbers to.
	 *
	 * @return True when all numbers before the precision digit is equal, false else-wise. Any
	 * number at or beyond the precision digit is ignored.
	 */
	[[nodiscard]]
	bool isEqual(
		const FloatingPoint& other,
		const std::uint8_t digits = defaultRoundingDigits
	) const;
	/**
	 * @brief Calculates the reciprocal of the current instance's value.
	 *
	 * @exception OverflowError When the result of the operation would result in an overflow.
	 * @exception UnderflowError When the result of the operation would result in an underflow.
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @return The value of 1 / value.
	 */
	[[nodiscard]]
	FloatingPoint reciprocal() const;
	/**
	 * @brief Converts the float-point value into a string.
	 *
	 * @return Human readable form of the number.
	 */
	[[nodiscard]]
	std::string toString() const;
	/**
	 * @brief Calculates the square root of the stored value.
	 *
	 * @exception DomainError When the result of the operation would be invalid.
	 *
	 * @return The result of the square root.
	 */
	[[nodiscard]]
	FloatingPoint squareRoot() const;
};
/**
 * @brief Represents the constituent parts of a real number.
 */
struct PECUNIA_EXPORT RealNumber
{
	/// The whole portion of a real number.
	FloatingPoint whole_;
	/// The fractional portion of a real number.
	FloatingPoint fraction_;
};
/**
 * @brief Relationship operator checking the less-than status.
 *
 * @param lhs The value to check is less-than the supplied right-hand-side value.
 * @param rhs The value to compare against.
 *
 * @return Gives true when the left-hand-side value is less-than the supplied right-hand-side value.
 */
[[nodiscard]]
PECUNIA_EXPORT bool operator<(const FloatingPointBase lhs, const FloatingPoint& rhs);
/**
 * @brief Relationship operator checking the greater-than status.
 *
 * @param lhs The value to check is greater-than the supplied right-hand-side value.
 * @param rhs The value to compare against.
 *
 * @return Gives true when the left-hand-side value is greater-than the supplied right-hand-side
 * value.
 */
[[nodiscard]]
PECUNIA_EXPORT bool operator>(const FloatingPointBase lhs, const FloatingPoint& rhs);
/**
 * @brief Calculates the multiplication operation.
 *
 * @exception OverflowError When the result of the operation would result in an overflow.
 * @exception UnderflowError When the result of the operation would result in an underflow.
 *
 * @param multiplicand The value that is to be multiplied.
 * @param multiplier The amount to multiply by.
 *
 * @return The result of the operation, the product.
 */
[[nodiscard]]
PECUNIA_EXPORT FloatingPoint operator*(
	const FloatingPointBase multiplicand,
	const FloatingPoint& multiplier
);
/**
 * @brief Calculates the division operation.
 *
 * @exception OverflowError When the result of the operation would result in an overflow.
 * @exception UnderflowError When the result of the operation would result in an underflow.
 * @exception DomainError When the result of the operation would result in a division by zero.
 *
 * @param dividend The value that is to be divided, sometimes called the numerator.
 * @param divisor The value to divide with, sometimes called the denominator.
 *
 * @return The result of the operation, the quotient.
 */
[[nodiscard]]
PECUNIA_EXPORT FloatingPoint operator/(
	const FloatingPointBase dividend,
	const FloatingPoint& divisor
);

}
namespace pecunia::literals
{

/**
 * @brief A numeric literal to create a FloatingPoint number.
 *
 * @param value value The base numeric value to convert.
 *
 * @return The converted number into the strong type representation.
 */
PECUNIA_EXPORT pecunia::FloatingPoint operator""_fp(long double value);

}

#endif
