/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Money.h"
using namespace std::string_literals;
using pecunia::FloatingPointBase;
using pecunia::FloatingPointRounderFunction;
using pecunia::MajorUnit;
using pecunia::MinorSign;
using pecunia::MinorUnit;
using pecunia::MoneyParts;
using pecunia::RounderFunction;
using pecunia::currency::Currency;
using pecunia::currency::Money;
using pecunia::currency::toIso4217Code;
using pecunia::currency::toStdString;
using pecunia::subMinorDigits;
using std::int16_t;
using std::int32_t;
using std::int64_t;
using std::int8_t;
using std::pair;
using std::string;
using std::to_string;
using std::uint8_t;

#include <cassert>
// Used for assert.
#include <cstdlib>
// Used for lldiv.
#include <cmath>
using std::abs;
using std::pow;
#include <exception>
using std::exception;
#include <iomanip>
using std::fixed;
using std::setfill;
using std::setprecision;
using std::setw;
#include <istream>
using std::istream;
#include <locale>
using std::locale;
using std::moneypunct;
using std::use_facet;
#include <limits>
using std::numeric_limits;
#include <ostream>
using std::ostream;
#include <sstream>
using std::ostringstream;

#include <internal/Adjustments.h>
using pecunia::currency::internal::fromAmountMajor;
using pecunia::currency::internal::fromAmountMinor;
using pecunia::currency::internal::fromAmountMinorOnly;
using pecunia::currency::internal::fromAmountSubMinor;
using pecunia::currency::internal::fromFloatingPointMajor;
using pecunia::currency::internal::fromFloatingPointMinor;
using pecunia::currency::internal::fromFloatingPointSubMinor;
using pecunia::currency::internal::fromMinorMinorOnly;
using pecunia::currency::internal::fromMinorSubMinor;
using pecunia::currency::internal::minorUnitSubMinorFactor;
using pecunia::currency::internal::normaliseAmount;
#include <internal/ManipIndex.h>
using pecunia::internal::countryLeadingIndex;
using pecunia::internal::useSymbolIndex;
using pecunia::internal::spacedIndex;
#include <internal/Sizes.hpp>
using pecunia::internal::extraFloatingPointPrecision;
#include <internal/Verification.h>
using pecunia::internal::VerifiedValue;
using pecunia::internal::verifyAdditionFits;
using pecunia::internal::verifyDivisionFits;
using pecunia::internal::verifyMultiplicationFits;
using pecunia::internal::verifySubtractionFits;
#include <pecunia/Errors.h>
using pecunia::DomainError;
using pecunia::LogicError;
using pecunia::OverflowError;
using pecunia::RuntimeError;
using pecunia::UnderflowError;
#include <pecunia/Rounders.h>
using pecunia::rounders::currency::none;

#include "FloatingPoint.h"
using namespace pecunia::literals;
using pecunia::FloatingPoint;
#include "Manip.h"
using pecunia::currency::spaced;
using pecunia::currency::trailing;


#define IMPLEMENT_MATH_OPERATOR(OtherType, op)  \
	Money Money::operator op (const OtherType& other) const  \
	{  \
		Money m{*this};  \
		m op ## = other;  \
		return m;  \
	}

#define IMPLEMENT_RELATIONAL_OPERATOR(OtherType, op)  \
	bool Money::operator op(const OtherType& value) const  \
	{  \
		return this->amount_ op value;  \
	}

#define IMPLEMENT_INEQUALITY_OPERATOR(OtherType)  \
	bool Money::operator!=(const OtherType& other) const  \
	{  \
		return ! (*this == other);  \
	}

#define IMPLEMENT_MATH_ASSIGNMENT_OPERATOR(OtherType, op)  \
	Money& Money::operator op (const OtherType& value)  \
	{  \
		return *this op static_cast<int64_t>(value);  \
	}

#define IMPLEMENT_MATH_OPERATOR_FUNCTION(LhsType, RhsType, op)  \
	Money pecunia::currency::operator op (const LhsType& lhs, const RhsType& rhs)  \
	{  \
		return rhs op lhs;  \
	}

MajorUnit Money::maximumMajorValue() const
{
	MajorUnit rawMaxMajor{
		static_cast<MajorUnit>(
			numeric_limits<MajorUnit>::max() - this->maximumMinorValue()
		)
	};
	const int32_t ratio{minorUnitSubMinorFactor(this->code_)};
	MajorUnit maxMajor{static_cast<MajorUnit>(rawMaxMajor / ratio)};
	return maxMajor;
}

MajorUnit Money::minimumMajorValue() const
{
	const auto maxMinorValue{this->maximumMinorValue()};
	MajorUnit rawMinMajor{
		static_cast<MajorUnit>(numeric_limits<MajorUnit>::lowest() + maxMinorValue)
	};
	const auto minorPrecisionRatio{minorUnitSubMinorFactor(this->code_)};
	MajorUnit minMajor{static_cast<MajorUnit>(rawMinMajor / minorPrecisionRatio)};
	return minMajor;
}

MinorUnit Money::maximumMinorValue() const
{
	const int32_t digits{minorUnitDigits(this->code_)};
	const MinorUnit max{static_cast<MinorUnit>(pow(10, digits + subMinorDigits) - 1.0)};
	return max;
}

MinorUnit Money::maximumMinorOnlyValue() const
{
	const int32_t digits{minorUnitDigits(this->code_)};
	const MinorUnit max{static_cast<MinorUnit>(pow(10, digits) - 1.0)};
	return max;
}

MinorUnit Money::maximumSubMinorValue() const
{
	const MinorUnit max{static_cast<MinorUnit>(pow(10, subMinorDigits) - 1.0)};
	return max;
}

MajorUnit Money::maximumAmountValue() const
{
	const int32_t unitRatio{minorUnitSubMinorFactor(this->code_)};
	const MajorUnit max{
		static_cast<MajorUnit>(
			(this->maximumMajorValue() * unitRatio) + this->maximumMinorValue()
		)
	};
	return max;
}

MajorUnit Money::minimumAmountValue() const
{
	const int32_t unitRatio{minorUnitSubMinorFactor(this->code_)};
	const auto minMajorValue{this->minimumMajorValue()};
	const auto maxMinorValue{this->maximumMinorValue()};
	const MajorUnit min{
		static_cast<MajorUnit>((minMajorValue * unitRatio) - maxMinorValue)
	};
	return min;
}

Money::Money() :
	Money{Currency::XXX}
{}

Money::Money(const Currency& code) :
	Money{code, 0u, 0u, 0u}
{}

Money::Money(
	const Currency& code,
	const MinorSign& sign,
	const MinorUnit& minor,
	const MinorUnit& subMinor
) :
	Money{code, 0u, minor, subMinor}
{
	assert(
		(minor > 0 || subMinor > 0 || (minor == 0 && subMinor == 0 && sign == MinorSign::Neither))
		&& "The value of zero has no sign."
	);

	if (sign == MinorSign::Negative)
		this->amount_ *= -1; // N.B. does not need the safe multiplication since it only contains
							// minor units.
}

Money::Money(
	const Currency& code,
	const MajorUnit& major,
	const MinorUnit& minor,
	const MinorUnit& subMinor
) :
	amount_{0},
	code_{code}
{
	assert(
		minorUnitSubMinorFactor(code) - 1.0 <= numeric_limits<MinorUnit>::max()
		&& "The number of minor digits required is too large and cannot fit inside the minor type."
	);
	const int32_t unitRatio{minorUnitSubMinorFactor(code)};

	if (major > this->maximumMajorValue())
		throw OverflowError{
			"The major money value, " + to_string(major)
				+ ", is too large to store within the currency."
		};

	if (major < this->minimumMajorValue())
		throw OverflowError{
			"The major money value, " + to_string(major)
				+ ", is too small to store within the currency."
		};

	if (minor > this->maximumMinorOnlyValue())
		throw OverflowError{
			"The minor money value, " + to_string(minor)
				+ ", is too large to store within the currency."
		};

	if (subMinor > this->maximumSubMinorValue())
		throw OverflowError{
			"The sub-minor money value, " + to_string(subMinor)
				+ ", is too large to store within the currency."
		};
	const auto adjustedMinor{
		static_cast<MinorUnit>(minor * pow(10, minorUnitDigits(code)))
	};

	if (major >= 0)
		this->amount_ = (major * unitRatio) + adjustedMinor + subMinor;
	else
		this->amount_ = (major * unitRatio) - adjustedMinor - subMinor;
}

Money::Money(
	const FloatingPointRounderFunction& rounder,
	const Currency& code,
	const FloatingPointBase& value
) :
	Money{
		code,
		fromFloatingPointMajor(value, code, subMinorDigits, rounder),
		fromFloatingPointMinor(value, code, subMinorDigits, rounder),
		fromFloatingPointSubMinor(value, code, subMinorDigits, rounder),
	}
{
	if (
		this->amount_ == 0
		&& ! FloatingPoint{value}.isEqual(
			0.0,
			static_cast<uint8_t>(
				minorUnitDigits(code) + subMinorDigits + extraFloatingPointPrecision
			)
		)
	)
		throw UnderflowError{
			"The value " + to_string(value)
				+ " is too small to store within the supplied precision."
		};
}

Money::Money(const MajorUnit amount, const Currency& code) :
	Money{
		code,
		fromAmountMajor(amount, code),
		fromAmountMinorOnly(amount, code),
		fromAmountSubMinor(amount, code)
	}
{}

Money::operator FloatingPointBase() const
{
	return this->toFloatingPoint(none);
}

Money Money::operator-() const
{
	Money m{*this};
	m *= -1;
	return m;
}

IMPLEMENT_MATH_OPERATOR(Money, +)

IMPLEMENT_MATH_OPERATOR(Money, -)

IMPLEMENT_MATH_OPERATOR(int8_t, *)

IMPLEMENT_MATH_OPERATOR(int16_t, *)

IMPLEMENT_MATH_OPERATOR(int32_t, *)

IMPLEMENT_MATH_OPERATOR(int64_t, *)

IMPLEMENT_MATH_OPERATOR(FloatingPointBase, *)

FloatingPointBase Money::operator/(const Money& other) const
{
	if (other.amount_ == 0)
		throw DomainError{
			"The division operation of " + to_string(this->amount_) + " and "
				+ to_string(other.amount_) + " result is undefined."
		};
	const FloatingPoint normalisedAmount{normaliseAmount(other, this->code_)};
	return (FloatingPoint{this->amount_} / normalisedAmount).underlying();
}

IMPLEMENT_MATH_OPERATOR(int8_t, /)

IMPLEMENT_MATH_OPERATOR(int16_t, /)

IMPLEMENT_MATH_OPERATOR(int32_t, /)

IMPLEMENT_MATH_OPERATOR(int64_t, /)

IMPLEMENT_MATH_OPERATOR(FloatingPointBase, /)

MajorUnit Money::operator%(const MajorUnit& value) const
{
	if (value == 0)
		throw DomainError{
			"The modulus operation of " + to_string(this->amount_) + " and "
			+ to_string(value) + " result is undefined."
		};
	return this->major() % value;
}

Money& Money::operator=(const Money& other)
{
	if (this != &other)
	{
		const MajorUnit normalisedAmount{normaliseAmount(other, this->code_)};
		const auto maximumAmountValue{this->maximumAmountValue()};
		const auto minimumAmountValue{this->minimumAmountValue()};

		if (normalisedAmount <= maximumAmountValue && normalisedAmount >= minimumAmountValue)
			this->amount_ = normalisedAmount;
		else
			throw OverflowError{
				"The assignment operation of " + to_string(other.amount_) + " to "
					+ to_string(this->amount_) + " is too large to fit within the precision."
			};
	}
	return *this;
}

bool Money::operator<(const Money& other) const
try // cppcheck-suppress syntaxError
{
	const MajorUnit normalisedAmount{normaliseAmount(other, this->code_)};
	return this->amount_ < normalisedAmount;
}
catch (const exception&)
{
	if (other.amount_ > 0) // other.amount_ is too large positive.
		return true;

	if (other.amount_ < 0) // other.amount_ is too large negative.
		return false;
	throw; // Something strange has occurred.
}

IMPLEMENT_RELATIONAL_OPERATOR(int8_t, <)

IMPLEMENT_RELATIONAL_OPERATOR(int16_t, <)

IMPLEMENT_RELATIONAL_OPERATOR(int32_t, <)

IMPLEMENT_RELATIONAL_OPERATOR(int64_t, <)

bool Money::operator<=(const Money& other) const
{
	return (*this < other) || (*this == other);
}

IMPLEMENT_RELATIONAL_OPERATOR(int8_t, <=)

IMPLEMENT_RELATIONAL_OPERATOR(int16_t, <=)

IMPLEMENT_RELATIONAL_OPERATOR(int32_t, <=)

IMPLEMENT_RELATIONAL_OPERATOR(int64_t, <=)

bool Money::operator>(const Money& other) const
{
	return other < *this;
}

IMPLEMENT_RELATIONAL_OPERATOR(int8_t, >)

IMPLEMENT_RELATIONAL_OPERATOR(int16_t, >)

IMPLEMENT_RELATIONAL_OPERATOR(int32_t, >)

IMPLEMENT_RELATIONAL_OPERATOR(int64_t, >)

bool Money::operator>=(const Money& other) const
{
	return (*this > other) || (*this == other);
}

IMPLEMENT_RELATIONAL_OPERATOR(int8_t, >=)

IMPLEMENT_RELATIONAL_OPERATOR(int16_t, >=)

IMPLEMENT_RELATIONAL_OPERATOR(int32_t, >=)

IMPLEMENT_RELATIONAL_OPERATOR(int64_t, >=)

bool Money::operator==(const Money& other) const
try
{
	const MajorUnit normalisedAmount{normaliseAmount(other, this->code_)};
	return this->amount_ == normalisedAmount;
}
catch (const exception&)
{
	// Doesn't matter why we threw, e.g. overflow, underflow, different codes, the error itself
	// is enough to know that the two values are not equal.
	return false;
}

IMPLEMENT_RELATIONAL_OPERATOR(int8_t, ==)

IMPLEMENT_RELATIONAL_OPERATOR(int16_t, ==)

IMPLEMENT_RELATIONAL_OPERATOR(int32_t, ==)

IMPLEMENT_RELATIONAL_OPERATOR(int64_t, ==)

IMPLEMENT_INEQUALITY_OPERATOR(Money)

IMPLEMENT_INEQUALITY_OPERATOR(int8_t)

IMPLEMENT_INEQUALITY_OPERATOR(int16_t)

IMPLEMENT_INEQUALITY_OPERATOR(int32_t)

IMPLEMENT_INEQUALITY_OPERATOR(int64_t)

Money& Money::operator+=(const Money& other)
{
	const MajorUnit normalisedAmount{normaliseAmount(other, this->code_)};
	const VerifiedValue verified{
		verifyAdditionFits(
			this->amount_,
			normalisedAmount,
			this->maximumAmountValue(),
			this->minimumAmountValue()
		)
	};
	switch (verified)
	{
	case VerifiedValue::Fits:
		this->amount_ += normalisedAmount;
		return *this;
	case VerifiedValue::Overflows:
		throw OverflowError{
			"The addition operation of " + to_string(this->amount_) + ' '
				+ toStdString(this->code_) + " and " + to_string(other.amount_) + ' '
				+ toStdString(other.code_) + " cannot be safely stored."
		};
	case VerifiedValue::Underflows:
		throw UnderflowError{
			"The addition operation of " + to_string(this->amount_) + ' '
				+ toStdString(this->code_) + " and " + to_string(other.amount_) + ' '
				+ toStdString(other.code_) + " cannot be safely stored."
		};
	case VerifiedValue::Undefined:
	default:
		throw LogicError{"Incorrect addition verification result."};
	};
}

Money& Money::operator-=(const Money& other)
{
	const MajorUnit normalisedAmount{normaliseAmount(other, this->code_)};
	const VerifiedValue verified{
		verifySubtractionFits(
			this->amount_,
			normalisedAmount,
			this->maximumAmountValue(),
			this->minimumAmountValue()
		)
	};
	switch (verified)
	{
	case VerifiedValue::Fits:
		this->amount_ -= normalisedAmount;
		return *this;
	case VerifiedValue::Overflows:
		throw OverflowError{
			"The subtraction operation of " + to_string(this->amount_) + ' '
				+ toStdString(this->code_) + " and " + to_string(other.amount_) + ' '
				+ toStdString(other.code_) + " cannot be safely stored."
		};
	case VerifiedValue::Underflows:
		throw UnderflowError{
			"The subtraction operation of " + to_string(this->amount_) + ' '
				+ toStdString(this->code_) + " and " + to_string(other.amount_) + ' '
				+ toStdString(other.code_) + " cannot be safely stored."
		};
	case VerifiedValue::Undefined:
	default:
		throw LogicError{"Incorrect subtraction verification result."};
	};
}

IMPLEMENT_MATH_ASSIGNMENT_OPERATOR(int8_t, *=)

IMPLEMENT_MATH_ASSIGNMENT_OPERATOR(int16_t, *=)

IMPLEMENT_MATH_ASSIGNMENT_OPERATOR(int32_t, *=)

Money& Money::operator*=(const int64_t& value)
{
	const VerifiedValue verified{
		verifyMultiplicationFits(
			this->amount_,
			value,
			this->maximumAmountValue(),
			this->minimumAmountValue()
		)
	};

	switch (verified)
	{
	case VerifiedValue::Fits:
		this->amount_ = this->amount_ * value;
		return *this;
	case VerifiedValue::Overflows:
		throw OverflowError{
			"The multiplication operation of " + to_string(this->amount_) + " and "
			+ to_string(value) + " cannot be safely stored."
		};
	case VerifiedValue::Underflows:
		throw UnderflowError{
			"The multiplication operation of " + to_string(this->amount_) + " and "
			+ to_string(value) + " cannot be safely stored."
		};
	case VerifiedValue::Undefined:
	default:
		throw LogicError{"Incorrect multiplication verification result."};
	};
}

Money& Money::operator*=(const FloatingPointBase& value)
{
	// Need to verify before adjustment and during the multiplication.
	const VerifiedValue verified{
		verifyMultiplicationFits(
			this->amount_,
			value,
			this->maximumAmountValue(),
			this->minimumAmountValue(),
			this->maximumMinorValue(),
			static_cast<uint8_t>(
				minorUnitDigits(this->code_) + subMinorDigits + extraFloatingPointPrecision
			)
		)
	};

	switch (verified)
	{
	case VerifiedValue::Fits:
		break;
	case VerifiedValue::Overflows:
		throw OverflowError{
			"The multiplication operation of " + to_string(this->amount_) + " and "
			+ to_string(value) + " cannot be safely stored."
		};
	case VerifiedValue::Underflows:
		throw UnderflowError{
			"The multiplication operation of " + to_string(this->amount_) + " and "
			+ to_string(value) + " cannot be safely stored."
		};
	case VerifiedValue::Undefined:
	default:
		throw LogicError{"Incorrect multiplication verification result."};
	};
	const auto minorRatio{minorUnitSubMinorFactor(this->code_)};
	const auto extraFpRatio{pow(10, extraFloatingPointPrecision)};
	const FloatingPoint unitRatio{minorRatio * extraFpRatio};
	const auto adjustedValue{FloatingPoint{value} * unitRatio};
	const auto preMultiplied{*this};

	try
	{
		*this *= static_cast<MajorUnit>(adjustedValue.underlying());
		this->amount_ /= static_cast<MajorUnit>(unitRatio.underlying());
	}
	catch (const exception&)
	{
		*this = preMultiplied;
		throw;
	}
	return *this;
}

IMPLEMENT_MATH_ASSIGNMENT_OPERATOR(int8_t, /=)

IMPLEMENT_MATH_ASSIGNMENT_OPERATOR(int16_t, /=)

IMPLEMENT_MATH_ASSIGNMENT_OPERATOR(int32_t, /=)

Money& Money::operator/=(const int64_t& value)
{
	const VerifiedValue verified{
		verifyDivisionFits(this->amount_, value, this->minimumAmountValue())
	};

	switch (verified)
	{
	case VerifiedValue::Fits:
		this->amount_ = this->amount_ / value;
		return *this;
	case VerifiedValue::Overflows:
		throw OverflowError{
			"The division operation of " + to_string(this->amount_) + " and "
				+ to_string(value) + " cannot be safely stored."
		};
	case VerifiedValue::Undefined:
		throw DomainError{
			"The division operation of " + to_string(this->amount_) + " and "
				+ to_string(value) + " result is undefined."
		};
	case VerifiedValue::Underflows:
		throw UnderflowError{
			"The division operation of " + to_string(this->amount_) + " and "
				+ to_string(value) + " cannot be safely stored."
		};
	default:
		throw LogicError{"Incorrect division verification result."};
	};
}

Money& Money::operator/=(const FloatingPointBase& value)
{
	// Need to verify before adjustment, during the multiplication, and division.
	const VerifiedValue preVerified{
		verifyDivisionFits(
			this->amount_,
			value,
			this->maximumAmountValue(),
			this->minimumAmountValue(),
			this->maximumMinorValue(),
			static_cast<uint8_t>(
				minorUnitDigits(this->code_) + subMinorDigits + extraFloatingPointPrecision
			)
		)
	};

	switch (preVerified)
	{
	case VerifiedValue::Fits:
		break;
	case VerifiedValue::Overflows:
		throw OverflowError{
			"The division operation of " + to_string(this->amount_) + " and "
				+ to_string(value) + " cannot be safely stored."
		};
	case VerifiedValue::Undefined:
		throw DomainError{
			"The division operation of " + to_string(this->amount_) + " and "
				+ to_string(value) + " result is undefined."
		};
	case VerifiedValue::Underflows:
		throw UnderflowError{
			"The division operation of " + to_string(this->amount_) + " and "
				+ to_string(value) + " cannot be safely stored."
		};
	default:
		throw LogicError{"Incorrect division verification result."};
	};
	const int64_t unitRatio{
		minorUnitSubMinorFactor(this->code_)
			* static_cast<int64_t>(pow(10, extraFloatingPointPrecision))
	};
	const VerifiedValue verifiedMultiplication{
		verifyMultiplicationFits(
			this->amount_,
			unitRatio,
			this->maximumAmountValue(),
			this->minimumAmountValue()
		)
	};

	switch (verifiedMultiplication)
	{
	case VerifiedValue::Fits:
		break;
	case VerifiedValue::Overflows:
		throw OverflowError{
			"The multiplication operation of " + to_string(this->amount_) + " and "
				+ to_string(value) + " cannot be safely stored during division."
		};
	case VerifiedValue::Underflows:
		throw UnderflowError{
			"The multiplication operation of " + to_string(this->amount_) + " and "
				+ to_string(value) + " cannot be safely stored during division."
		};
	case VerifiedValue::Undefined:
	default:
		throw LogicError{"Incorrect multiplication verification result during division."};
	};
	const auto adjustedValue{
		static_cast<MajorUnit>(
			(FloatingPoint{value} * FloatingPoint{unitRatio}).underlying()
		)
	};
	const auto adjustedAmount{this->amount_ * unitRatio};
	const auto preAmount{this->amount_};

	try
	{
		this->amount_ = adjustedAmount;
		*this /= adjustedValue;
		return *this;
	}
	catch (const exception&)
	{
		this->amount_ = preAmount;
		throw;
	}
}

const MajorUnit& Money::amount() const noexcept
{
	return this->amount_;
}

MajorUnit Money::major() const
{
	return fromAmountMajor(this->amount_, this->code_);
}

MinorUnit Money::minor() const
{
	return fromAmountMinor(this->amount_, this->code_);
}

MinorUnit Money::minorOnly() const
{
	return fromAmountMinorOnly(this->amount_, this->code_);
}

MinorUnit Money::subMinor() const
{
	return fromAmountSubMinor(this->amount_, this->code_);
}

const Currency& Money::code() const noexcept
{
	return this->code_;
}

FloatingPointBase Money::toFloatingPoint(const RounderFunction& rounder, const uint8_t digits) const
{
	assert(rounder != nullptr && "The money rounder must be a function.");
	const auto minorDigits{
		static_cast<uint8_t>(minorUnitDigits(this->code_) + subMinorDigits)
	};
	const FloatingPoint rounded{rounder(this->amount_, minorDigits, digits)};
	const FloatingPoint ratio{minorUnitSubMinorFactor(this->code_)};
	const FloatingPoint converted{rounded / ratio};
	return converted.underlying();
}

Money Money::round(const RounderFunction& rounder, const uint8_t digits) const
{
	assert(rounder != nullptr && "The money rounder must be a function.");
	const auto minorDigits{
		static_cast<uint8_t>(minorUnitDigits(this->code_) + subMinorDigits)
	};
	const auto roundedAmount{rounder(this->amount_, minorDigits, digits)};
	Money rounded{roundedAmount, this->code_};
	return rounded;
}

bool Money::hasAmount() const noexcept
{
	return this->amount_ != 0 && this->code_ != Currency::XXX;
}

MoneyParts Money::maximum(const Currency& code) noexcept
{
	const Money test{code};
	const auto amount{test.maximumAmountValue()};
	const auto majorMinorFactor{minorUnitSubMinorFactor(code)};
	const MajorUnit major{amount / majorMinorFactor};
	const auto minor{static_cast<MinorUnit>(amount % majorMinorFactor)};
	const auto subMinorFactor{static_cast<MinorUnit>(pow(10, subMinorDigits))};
	const auto minorOnly{static_cast<MinorUnit>(minor / subMinorFactor)};
	const auto subMinor{static_cast<MinorUnit>(minor % subMinorFactor)};
	return {amount, major, minor, minorOnly, subMinor, code};
}

MoneyParts Money::minimum(const Currency& code) noexcept
{
	const Money test{code};
	const auto amount{test.minimumAmountValue()};
	const auto majorMinorFactor{minorUnitSubMinorFactor(code)};
	const MajorUnit major{amount / majorMinorFactor};
	const auto minor{static_cast<MinorUnit>(-amount % majorMinorFactor)};
	const auto subMinorFactor{static_cast<MinorUnit>(pow(10, subMinorDigits))};
	const auto minorOnly{static_cast<MinorUnit>(minor / subMinorFactor)};
	const auto subMinor{static_cast<MinorUnit>(minor % subMinorFactor)};
	return {amount, major, minor, minorOnly, subMinor, code};
}

void Money::zeroOut()
{
	this->amount_ = 0;
}

bool Money::hasOnlyMinor() const noexcept
{
	const int32_t digits{minorUnitDigits(this->code_)};
	const MinorUnit majorPoint{static_cast<MinorUnit>(pow(10, digits + subMinorDigits))};
	return this->amount_ > -majorPoint && this->amount_ < majorPoint;
}

Money& Money::assign(const Money& other) noexcept
{
	this->code_ = other.code_;
	this->amount_ = other.amount_;
	return *this;
}

IMPLEMENT_MATH_OPERATOR_FUNCTION(int8_t, Money, *)

IMPLEMENT_MATH_OPERATOR_FUNCTION(int16_t, Money, *)

IMPLEMENT_MATH_OPERATOR_FUNCTION(int32_t, Money, *)

IMPLEMENT_MATH_OPERATOR_FUNCTION(int64_t, Money, *)

IMPLEMENT_MATH_OPERATOR_FUNCTION(FloatingPointBase, Money, *)

ostream& pecunia::currency::operator<<(ostream& stream, const Money& m)
{
	const bool isSymbolUsed{stream.iword(useSymbolIndex()) == 1};
	const string currencyMarker{
		isSymbolUsed ? currencySymbol(m.code_).value_or(toStdString(m.code_)) : toStdString(m.code_)
	};
	stream.exceptions(istream::failbit | istream::badbit);
	const bool isCountryLeading{stream.iword(countryLeadingIndex()) == 1};
	const bool isSpaced{stream.iword(spacedIndex()) == 1};
	stream << (isCountryLeading ? currencyMarker : "")
			<< (isSpaced && isCountryLeading ? " " : "")
			<< m.major()
			<< use_facet<moneypunct<char>>(stream.getloc()).decimal_point()
			<< setw(minorUnitDigits(m.code_) + subMinorDigits) << setfill('0')
			<< m.minor()
			<< (isSpaced && ! isCountryLeading ? " " : "")
			<< ( ! isCountryLeading ? currencyMarker : "");
	return stream;
}

istream& pecunia::currency::operator>>(istream& stream, Money& m)
try
{
	MajorUnit major;
	MinorUnit minor;
	char separator;
	string rawCode;
	stream.exceptions(istream::failbit | istream::badbit);
	const bool isCountryLeading{stream.iword(countryLeadingIndex()) == 1};

	if (isCountryLeading)
		stream >> rawCode >> major >> separator >> minor;
	else
		stream >> major >> separator >> minor >> rawCode;

	if (separator != use_facet<moneypunct<char>>(stream.getloc()).decimal_point())
		throw RuntimeError{"The currency is incorrectly formatted."};
	m.code_ = toIso4217Code(rawCode);
	const auto minorOnly{fromMinorMinorOnly(minor, subMinorDigits)};
	const auto subMinor{fromMinorSubMinor(minor, subMinorDigits)};
	m = Money{m.code_, major, minorOnly, subMinor};
	return stream;
}
catch (const Error&)
{
	throw;
}
catch (const exception& error)
{
	throw RuntimeError{"Failure reading money.", error};
}

string pecunia::currency::toStdString(const Money& m)
{
	ostringstream oss;
	oss << spaced << trailing << m;
	return oss.str();
}

void pecunia::currency::swap(Money& lhs, Money& rhs) noexcept
{
	const Money temp{lhs};
	lhs.amount_ = rhs.amount_;
	lhs.code_ = rhs.code_;
	rhs.amount_ = temp.amount_;
	rhs.code_ = temp.code_;
}

#undef IMPLEMENT_MATH_OPERATOR
#undef IMPLEMENT_RELATIONAL_OPERATOR
#undef IMPLEMENT_NOT_EQUAL_OPERATOR
#undef IMPLEMENT_MATH_ASSIGNMENT_OPERATOR
#undef IMPLEMENT_MATH_OPERATOR_FUNCTION
