/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_MONEY_H
#define PECUNIA_MONEY_H

#include <cstdint>
#include <functional>
#include <iosfwd>
#include <string>
#include <utility>

#include "Codes.h"
#include "Types.h"


namespace pecunia
{

class FloatingPoint;

}
namespace pecunia::currency
{

class Money;

}
namespace pecunia::math
{

PECUNIA_EXPORT currency::Money absoluteValue(const currency::Money&);

}
namespace pecunia::currency
{

/**
 * @brief A representation of a country's currency supporting basic arithmetic usage and
 * guaranteeing all operations succeed without a loss of precision, overflow, or underflow.
 */
class PECUNIA_EXPORT Money
{
	friend PECUNIA_EXPORT Money math::absoluteValue(const Money&);
	friend PECUNIA_EXPORT Money operator*(const std::int8_t& lhs, const Money& rhs);
	friend PECUNIA_EXPORT Money operator*(const std::int16_t& lhs, const Money& rhs);
	friend PECUNIA_EXPORT Money operator*(const std::int32_t& lhs, const Money& rhs);
	friend PECUNIA_EXPORT Money operator*(const std::int64_t& lhs, const Money& rhs);
	friend PECUNIA_EXPORT Money operator*(const FloatingPointBase& lhs, const Money& rhs);
	friend PECUNIA_EXPORT std::ostream& operator<<(std::ostream& stream, const Money& m);
	friend PECUNIA_EXPORT std::istream& operator>>(std::istream& stream, Money& m);
	friend PECUNIA_EXPORT void swap(Money& lhs, Money& rhs) noexcept;

	/**
	 * @brief Calculates the largest storable monetary major value.
	 */
	[[nodiscard]]
	MajorUnit maximumMajorValue() const;
	/**
	 * @brief Calculates the smallest storable monetary major value.
	 */
	[[nodiscard]]
	MajorUnit minimumMajorValue() const;
	/**
	 * @brief Calculates the largest storable monetary minor value, i.e. accounts for a currency's
	 * minor digits and sub-minor digits.
	 */
	[[nodiscard]]
	MinorUnit maximumMinorValue() const;
	/**
	 * @brief Calculates the largest storable monetary truncated minor value, i.e. only a currency's
	 * minor digits.
	 */
	[[nodiscard]]
	MinorUnit maximumMinorOnlyValue() const;
	/**
	 * @brief Calculates the largest storable monetary sub-minor value.
	 */
	[[nodiscard]]
	MinorUnit maximumSubMinorValue() const;
	/**
	 * @brief Calculates the largest storable monetary amount value.
	 */
	[[nodiscard]]
	MajorUnit maximumAmountValue() const;
	/**
	 * @brief Calculates the smallest storable monetary amount value.
	 */
	[[nodiscard]]
	MajorUnit minimumAmountValue() const;

	/// The entire representation of the monetary amount using the currency's minor unit with the
	/// extra sub-minor digits.
	MajorUnit amount_;
	/// The currency code the monetary amount represents.
	Currency code_;

public:
	/**
	 * @brief Constructor for a fully formed object that initialises the stored amount to an
	 * unknown currency at zero.
	 */
	Money();
	/**
	 * @brief Constructor for a fully formed object that initialises the stored amount to zero of a
	 * specific currency.
	 *
	 * @param[in] code The currency code the amount represents.
	 */
	explicit Money(const Currency& code);
	/**
	 * @brief Constructor for setting an exact minor monetary value whose major value is zero.
	 *
	 * @pre When the minor value is zero, the sign must be MinorSign::Neither.
	 *
	 * @exception OverflowError When the minor unit is too large to be stored.
	 *
	 * @param[in] code The currency code the amount represents.
	 * @param[in] sign The associated sign of the minor value, e.g. negative.
	 * @param[in] minor The amount of the minor currency to store, e.g. U.S.A. cents.
	 * @param[in] minor The amount of the sub-minor currency to store, e.g. U.S.A. fractional cents.
	 */
	Money(
		const Currency& code,
		const MinorSign& sign,
		const MinorUnit& minor,
		const MinorUnit& subMinor = 0u
	);
	/**
	 * @brief Constructor for setting an exact monetary value of a fully formed object.
	 *
	 * @pre The maximum minor value must fit within the MinorUnit.
	 *
	 * @exception OverflowError When the major unit is too large to be stored in a positive
	 * direction.
	 * @exception OverflowError When the major unit is too small to be stored in a negative
	 * direction.
	 * @exception OverflowError When the minor unit is too large to be stored.
	 *
	 * @param[in] code The currency code the amount represents.
	 * @param[in] major The amount of major currency to store, e.g. U.S.A. dollars.
	 * @param[in] minor The amount of the minor currency to store, e.g. U.S.A. cents.
	 * @param[in] minor The amount of the sub-minor currency to store, e.g. U.S.A. fractional cents.
	 */
	Money(
		const Currency& code,
		const MajorUnit& major,
		const MinorUnit& minor = 0,
		const MinorUnit& subMinor = 0
	);
	/**
	 * @brief Constructor for setting a monetary value using the native floating-point type.
	 *
	 * @note The minor units in a floating-point value cannot be verified and will always fit within
	 * the currency digits and precision specified, e.g. U.S.A. dollar value with a sub-minor digits
	 * of 0 will store the value 1.555 as 1.55 when no rounder is supplied.
	 *
	 * @exception OverflowError When the major unit is too large to be stored in a positive
	 * direction.
	 * @exception OverflowError When the major unit is too small to be stored in a negative
	 * direction.
	 * @exception DomainError When the conversion results in the floating-point value dividing by
	 * zero.
	 * @exception UnderflowError When the floating-point value is too small to fit within the
	 * FloatingPoint.
	 * @exception OverflowError When the floating-point value is too large to fit within the
	 * FloatingPoint.
	 * @exception DomainError When the conversion results in the floating-point value not being
	 * valid.
	 *
	 * @param[in] code The currency code the amount represents.
	 * @param[in] value The floating-point value to convert.
	 * @param[in] rounder The method for rounding the number when converting the floating-point
	 * number.
	 */
	Money(
		const FloatingPointRounderFunction& rounder,
		const Currency& code,
		const FloatingPointBase& value
	);
	/**
	 * @brief Full initialisation constructor from a stored amount.
	 *
	 * @exception OverflowError When the major unit is too large to be stored in a positive
	 * direction.
	 * @exception OverflowError When the major unit is too small to be stored in a negative
	 * direction.
	 * @exception OverflowError When the minor unit is too large to be stored.
	 *
	 * @param amount The entire representation of the monetary amount using the currency's major
	 * unit, minor unit, and the sub-minor digits, in the form MajorMinorSub-minor.
	 * @param code The ISO currency code of the amount.
	 */
	Money(const MajorUnit amount, const Currency& code);
	Money(const Money&) =default;
	Money(Money&&) noexcept =default;
	/**
	 * @brief Conversion operator to a native floating-point type.
	 *
	 * @exception DomainError When the conversion results in the floating-point value dividing by
	 * zero.
	 * @exception UnderflowError When the floating-point value is too small to fit within the
	 * FloatingPoint.
	 * @exception OverflowError When the floating-point value is too large to fit within the
	 * FloatingPoint.
	 * @exception DomainError When the conversion results in the floating-point value not being
	 * valid.
	 */
	explicit operator FloatingPointBase() const;
	~Money() =default;
	/**
	 * @brief Negation operator without any loss of precision during the conversion.
	 *
	 * @return The negated value of the current object.
	 */
	[[nodiscard]]
	Money operator-() const;
	/**
	 * @brief Additive operator. When the supplied monetary object is of a different currency, the
	 * amount is converted into the current object's currency using the pecunia::currency::converter
	 * function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] other The additional value to add to the current object's value.
	 *
	 * @return A new object containing the summation of the two values.
	 */
	[[nodiscard]]
	Money operator+(const Money& other) const;
	/**
	 * @brief Subtractive operator. When the supplied monetary object is of a different currency,
	 * the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] other The additional value to subtract from the current object's value.
	 *
	 * @return A new object containing the difference of the two values.
	 */
	[[nodiscard]]
	Money operator-(const Money& other) const;
	Money operator*(const Money&) const =delete;
	/**
	 * @brief Multiplicative operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The signed 8-bit integer value to multiply to the current object's value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const std::int8_t& value) const;
	/**
	 * @brief Multiplicative operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The signed 16-bit integer value to multiply to the current object's value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const std::int16_t& value) const;
	/**
	 * @brief Multiplicative operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The signed 32-bit integer value to multiply to the current object's value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const std::int32_t& value) const;
	/**
	 * @brief Multiplicative operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The signed 64-bit integer value to multiply to the current object's value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const std::int64_t& value) const;
	/**
	 * @brief Multiplicative operator without any loss of precision. This only supports as many
	 * digits as there are minor digits.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional value to multiply to the current object's value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const FloatingPointBase& value) const;
	/**
	 * @brief Division operator for calculating ratios.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @param[in] other The value to divide by the current object's value.
	 *
	 * @return The result of the division between the two, a ratio.
	 */
	[[nodiscard]]
	FloatingPointBase operator/(const Money& other) const;
	/**
	 * @brief Division operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional signed 8-bit integer value to divide by the current object's
	 * value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const std::int8_t& value) const;
	/**
	 * @brief Division operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional signed 16-bit integer value to divide by the current object's
	 * value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const std::int16_t& value) const;
	/**
	 * @brief Division operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional signed 32-bit integer value to divide by the current object's
	 * value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const std::int32_t& value) const;
	/**
	 * @brief Division operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional signed 64-bit integer value to divide by the current object's
	 * value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const std::int64_t& value) const;
	/**
	 * @brief Division operator without any loss of precision. This only supports as many digits as
	 * there are minor digits.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional value to divide by the current object's value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const FloatingPointBase& value) const;
	Money operator%(const Money&) const =delete;
	/**
	 * @brief Modulus operator performed upon /only/ the major units of the object's current value.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @param[in] value The additional value to perform a "mod" on the current object's value.
	 *
	 * @return The value of the modulus operation of the two values.
	 */
	[[nodiscard]]
	MajorUnit operator%(const MajorUnit& value) const;
	/**
	 * @brief Assignment Operator with copy. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception OverflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] other The object whose amount is to be assigned into the current object.
	 *
	 * @post The current object's amount is changed to the supplied one. The original object's
	 * currency does not change.
	 *
	 * @return The current object with the new amount.
	 */
	Money& operator=(const Money& other);
	/**
	 * @brief Less-than Relational Operator. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @param[in] other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const Money& other) const;
	/**
	 * @brief Less-than Relational Operator for an 8-bit signed integer. The supplied value is
	 * expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const std::int8_t& value) const;
	/**
	 * @brief Less-than Relational Operator for an 16-bit signed integer. The supplied value is
	 * expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const std::int16_t& value) const;
	/**
	 * @brief Less-than Relational Operator for an 32-bit signed integer. The supplied value is
	 * expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const std::int32_t& value) const;
	/**
	 * @brief Less-than Relational Operator for an 64-bit signed integer. The supplied value is
	 * expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const std::int64_t& value) const;
	/**
	 * @brief Less-than-equal Relational Operator. When the supplied monetary object is of a
	 * different currency, the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @param[in] other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const Money& other) const;
	/**
	 * @brief Less-than-equal Relational Operator for an 8-bit signed integer. The supplied value is
	 * expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const std::int8_t& value) const;
	/**
	 * @brief Less-than-equal Relational Operator for an 16-bit signed integer. The supplied value
	 * is expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const std::int16_t& value) const;
	/**
	 * @brief Less-than-equal Relational Operator for an 32-bit signed integer. The supplied value
	 * is expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const std::int32_t& value) const;
	/**
	 * @brief Less-than-equal Relational Operator for an 64-bit signed integer. The supplied value
	 * is expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const std::int64_t& value) const;
	/**
	 * @brief Greater-than Relational Operator. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @param[in] other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const Money& other) const;
	/**
	 * @brief Greater-than Relational Operator for an 8-bit signed integer. The supplied value is
	 * expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const std::int8_t& value) const;
	/**
	 * @brief Greater-than Relational Operator for an 16-bit signed integer. The supplied value is
	 * expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const std::int16_t& value) const;
	/**
	 * @brief Greater-than Relational Operator for an 32-bit signed integer. The supplied value is
	 * expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const std::int32_t& value) const;
	/**
	 * @brief Greater-than Relational Operator for an 64-bit signed integer. The supplied value is
	 * expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const std::int64_t& value) const;
	/**
	 * @brief Greater-than-equal Relational Operator. When the supplied monetary object is of a
	 * different currency, the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @param[in] other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const Money& other) const;
	/**
	 * @brief Greater-than-equal Relational Operator for an 8-bit signed integer. The supplied value
	 * is expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const std::int8_t& value) const;
	/**
	 * @brief Greater-than-equal Relational Operator for an 16-bit signed integer. The supplied
	 * value is expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const std::int16_t& value) const;
	/**
	 * @brief Greater-than-equal Relational Operator for an 32-bit signed integer. The supplied
	 * value is expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const std::int32_t& value) const;
	/**
	 * @brief Greater-than-equal Relational Operator for an 64-bit signed integer. The supplied
	 * value is expected to be expressed in terms of the current object's issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const std::int64_t& value) const;
	/**
	 * @brief Equality Operator. When the supplied monetary object is of a different currency,
	 * the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @param[in] other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const Money& other) const;
	/**
	 * @brief Equality Operator for an 8-bit signed integer. The supplied value is expected to be
	 * expressed in terms of the current object's issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied value within, the
	 * expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const std::int8_t& value) const;
	/**
	 * @brief Equality Operator for an 16-bit signed integer. The supplied value is expected to be
	 * expressed in terms of the current object's issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied value within, the
	 * expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const std::int16_t& value) const;
	/**
	 * @brief Equality Operator for an 32-bit signed integer. The supplied value is expected to be
	 * expressed in terms of the current object's issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied value within, the
	 * expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const std::int32_t& value) const;
	/**
	 * @brief Equality Operator for an 64-bit signed integer. The supplied value is expected to be
	 * expressed in terms of the current object's issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied value within, the
	 * expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const std::int64_t& value) const;
	/**
	 * @brief Inequality Operator. When the supplied monetary object is of a different currency,
	 * the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @param[in] other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const Money& other) const;
	/**
	 * @brief Inequality Operator for an 8-bit signed integer. The supplied value is expected to be
	 * expressed in terms of the current object's issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied value, within
	 * the expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const std::int8_t& value) const;
	/**
	 * @brief Inequality Operator for an 16-bit signed integer. The supplied value is expected to be
	 * expressed in terms of the current object's issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied value, within
	 * the expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const std::int16_t& value) const;
	/**
	 * @brief Inequality Operator for an 32-bit signed integer. The supplied value is expected to be
	 * expressed in terms of the current object's issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied value, within
	 * the expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const std::int32_t& value) const;
	/**
	 * @brief Inequality Operator for an 64-bit signed integer. The supplied value is expected to be
	 * expressed in terms of the current object's issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param[in] value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied value, within
	 * the expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const std::int64_t& value) const;
	/**
	 * @brief Additive assignment operator. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] other The additional value to add to the current object.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the summation of the two values.
	 */
	Money& operator+=(const Money& other);
	/**
	 * @brief Subtractive assignment operator. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object's currency using the
	 * pecunia::currency::converter function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] other The additional value to subtract from the current object.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the difference of the two values.
	 */
	Money& operator-=(const Money& other);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional signed 8-bit integer value to multiply to the current
	 * object.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const std::int8_t& value);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional signed 16-bit integer value to multiply to the current
	 * object.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const std::int16_t& value);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional signed 32-bit integer value to multiply to the current
	 * object.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const std::int32_t& value);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional signed 64-bit integer value to multiply to the current
	 * object.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const std::int64_t& value);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision. This only supports
	 * as many digits as there are minor digits.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional value to multiply to the current object.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const FloatingPointBase& value);
	/**
	 * @brief Division assignment operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The signed 8-bit integer additional value to divide by the current object's
	 * value.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const std::int8_t& value);
	/**
	 * @brief Division assignment operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The signed 16-bit integer additional value to divide by the current object's
	 * value.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const std::int16_t& value);
	/**
	 * @brief Division assignment operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The signed 32-bit integer additional value to divide by the current object's
	 * value.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const std::int32_t& value);
	/**
	 * @brief Division assignment operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The signed 64-bit integer additional value to divide by the current object's
	 * value.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const std::int64_t& value);
	/**
	 * @brief Division assignment operator without any loss of precision. This only supports as many
	 * digits as there are minor digits.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param[in] value The additional value to divide by the current object's value.
	 *
	 * @post The current object's amount is changed using the mathematical operation. The current
	 * object's currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const FloatingPointBase& value);
	/**
	 * @brief The entire representation of the monetary amount using the currency's major unit,
	 * minor unit, and the sub-minor digits, in the form MajorMinorSub-minor.
	 */
	[[nodiscard]]
	const MajorUnit& amount() const noexcept;
	/**
	 * @brief Accessor to the major monetary value, e.g. only the U.S.A. dollars.
	 */
	[[nodiscard]]
	MajorUnit major() const;
	/**
	 * @brief Accessor to the minor and sub-minor monetary value, e.g. the U.S.A. cents and
	 * fractional cents.
	 */
	[[nodiscard]]
	MinorUnit minor() const;
	/**
	 * @brief Accessor to the minor monetary value, e.g. the U.S.A. cents, truncating the fractional
	 * cents.
	 */
	[[nodiscard]]
	MinorUnit minorOnly() const;
	/**
	 * @brief Accessor to the sub-minor monetary value, e.g. only the U.S.A. fractional cents.
	 */
	[[nodiscard]]
	MinorUnit subMinor() const;
	/**
	 * @brief Accessor to the code the currency amount represents.
	 */
	[[nodiscard]]
	const Currency& code() const noexcept;
	/**
	 * @brief Converts the monetary value into a native floating-point type using a supplied
	 * rounding method.
	 *
	 * @param[in] rounder The rounding function to apply to the monetary value.
	 * @param[in] digits  The number of digits after the major unit.
	 *
	 * @exception DomainError When the conversion results in the floating-point value dividing by
	 * zero.
	 * @exception UnderflowError When the floating-point value is too small to fit within the
	 * FloatingPoint.
	 * @exception OverflowError When the floating-point value is too large to fit within the
	 * FloatingPoint.
	 * @exception DomainError When the conversion results in the floating-point value not being
	 * valid.
	 */
	[[nodiscard]]
	FloatingPointBase toFloatingPoint(
		const RounderFunction& rounder,
		const std::uint8_t digits = defaultRoundingDigits
	) const;
	/**
	 * @brief Performs a rounding operation upon the monetary value.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception OverflowError When the result of the operation would be too small to store.
	 * @exception DomainError When the conversion results in the floating-point value dividing by
	 * zero.
	 * @exception UnderflowError When the floating-point value is too small to fit within the
	 * FloatingPoint.
	 * @exception OverflowError When the floating-point value is too large to fit within the
	 * FloatingPoint.
	 * @exception DomainError When the conversion results in the floating-point value not being
	 * valid.
	 *
	 * @param rounder The rounding function to apply to the monetary value.
	 * @param digits The number of digits after the major unit to round to.
	 *
	 * @return A copy of the current object containing the rounded value.
	 */
	[[nodiscard]]
	Money round(
		const RounderFunction& rounder,
		const std::uint8_t digits = defaultRoundingDigits
	) const;
	/**
	 * @brief Determines if the instance has an amount set. For example, an amount of 2.34 USD would
	 * be set, but 0.00 XXX would not be.
	 *
	 * @return When the major, minor, sub-minor, and code are all set gives true, false else-wise.
	 */
	[[nodiscard]]
	bool hasAmount() const noexcept;
	/**
	 * @brief Calculates the largest monetary value that can be used for a given currency.
	 *
	 * @param code The currency code to calculate the monetary value in.
	 *
	 * @return The monetary components of the maximum value.
	 */
	[[nodiscard]]
	static MoneyParts maximum(const Currency& code) noexcept;
	/**
	 * @brief Calculates the smallest monetary value that can be used for a given currency.
	 *
	 * @param code The currency code to calculate the monetary value in.
	 *
	 * @return The monetary components of the minimum value.
	 */
	[[nodiscard]]
	static MoneyParts minimum(const Currency& code) noexcept;
	/**
	 * @brief Resets the monetary value to zero, without changing the currency.
	 */
	void zeroOut();
	/**
	 * @brief Determines if the monetary amount consists of only minor units, i.e. only change.
	 *
	 * @return Gives true when the amount is exclusively between negative one major unit and one
	 * major unit, false otherwise.
	 */
	[[nodiscard]]
	bool hasOnlyMinor() const noexcept;
	/**
	 * @brief An assignment operation using the supplied value.
	 *
	 * @param other The value to assign.
	 *
	 * @post The current object's currency and amount is changed to the supplied one.
	 *
	 * @return The current object after assignment.
	 */
	Money& assign(const Money& other) noexcept;
};

/**
 * @brief Multiplicative operator without any loss of precision.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param[in] lhs The signed 8-bit integer amount by which to multiply the supplied object's
 * value.
 * @param[in] rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const std::int8_t& lhs, const Money& rhs);
/**
 * @brief Multiplicative operator without any loss of precision.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param[in] lhs The signed 16-bit integer amount by which to multiply the supplied object's
 * value.
 * @param[in] rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const std::int16_t& lhs, const Money& rhs);
/**
 * @brief Multiplicative operator without any loss of precision.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param[in] lhs The signed 32-bit integer amount by which to multiply the supplied object's
 * value.
 * @param[in] rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const std::int32_t& lhs, const Money& rhs);
/**
 * @brief Multiplicative operator without any loss of precision.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param[in] lhs The signed 64-bit integer amount by which to multiply the supplied object's
 * value.
 * @param[in] rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const std::int64_t& lhs, const Money& rhs);
/**
 * @brief Multiplicative operator without any loss of precision. This only supports as many digits
 * as there are minor digits.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param[in] lhs The amount by which to multiply the supplied object's value.
 * @param[in] rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const FloatingPointBase& lhs, const Money& rhs);
/**
 * @brief Stream Insertion Operator. By default the form will be the locale's monetary format with
 * no spaces, the ISO code, and trailing the value, e.g. a U.S.A. locale with value of one dollar
 * and twenty-three cents would appear as 1.23USD.
 *
 * @exception RuntimeError When the stream fails to be written.
 *
 * @param[in,out] stream The destination source for the contents of the supplied money object.
 * @param[in] m The object whose values are to placed into the destination source.
 *
 * @return The modified source stream object filled with the contents of the money object.
 */
PECUNIA_EXPORT std::ostream& operator<<(std::ostream& stream, const Money& m);
/**
 * @brief Stream Extraction Operator. The monetary amount will be read in using the locale's setting
 * for reading in money. The currency code is required and must be in all upper-case letters.
 *
 * @pre The locale of the supplied stream must be set to get the money formatted in readable
 * manner.
 *
 * @exception RuntimeError When the stream fails to be read.
 * @exception InvalidArgument When the currency is incorrectly formatted.
 *
 * @param[in,out] stream The extraction source for the contents of the supplied money object.
 * @param[out] m The object whose values are to be filled by the contents of the stream.
 *
 * @return The modified source stream object emptied of a money entry.
 */
PECUNIA_EXPORT std::istream& operator>>(std::istream& stream, Money& m);
/**
 * @brief Converts the monetary representation into a string from. The format of the string
 * representation is the amount followed by a space and the currency code.
 *
 * @param[out] m The object whose values are to be converted into a string form.
 *
 * @return The string-fied representation of the monetary amount.
 */
[[nodiscard]]
PECUNIA_EXPORT std::string toStdString(const Money& m);
/**
 * @brief Exchanges two monetary objects without any conversion operation performed.
 *
 * @param lhs The object to swap to the right.
 * @param rhs The object to swap to the left.
 */
[[nodiscard]]
PECUNIA_EXPORT void swap(Money& lhs, Money& rhs) noexcept;

}

#endif
