/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Rounders.h"
using pecunia::MajorUnit;
using pecunia::MinorUnit;
using std::uint8_t;

#include <cassert>
// Used for assert.
#include <cmath>
using std::abs;
using std::ceil;
using std::floor;
using std::fmod;
using std::modf;
using std::pow;
#include <functional>
using std::function;
#include <string>
using std::to_string;

#include <pecunia/Errors.h>
using pecunia::InvalidArgument;
using pecunia::LogicError;
using pecunia::RuntimeError;
#include <pecunia/FloatingPoint.h>
using namespace pecunia::literals;
using pecunia::FloatingPoint;


namespace
{

enum class PointRoundingDirection
{
	None,
	Up,
	Down,
};

FloatingPoint round(
	const FloatingPoint& number,
	const uint8_t& digitsLeft,
	const function<
								   // roundingDigit,          wholeNumber, isPositive)
		PointRoundingDirection(const FloatingPoint&, const FloatingPoint&, const bool)
	>& selectDirection
)
{
	const FloatingPoint shiftOver{10.0_fp .power(digitsLeft)};
	const auto splitNumber{number.decompose(digitsLeft)};
	const auto shiftedNumberFraction{splitNumber.fraction_ * shiftOver};
	const auto splitRoundPoint{shiftedNumberFraction.decompose(digitsLeft)};
	const bool isPositive{number > 0};
	const auto direction{
		selectDirection(splitRoundPoint.fraction_.absoluteValue(), splitNumber.whole_, isPositive)
	};
	const auto roundedFractionPoint{
		[&] () -> FloatingPoint
		{
			switch (direction)
			{
			case PointRoundingDirection::Up:
				return splitRoundPoint.fraction_
					+ (
						isPositive
							? 1.0_fp - splitRoundPoint.fraction_
							: -splitRoundPoint.fraction_
					);
			case PointRoundingDirection::Down:
				return splitRoundPoint.fraction_
					- (
						isPositive
							? splitRoundPoint.fraction_
							: 1.0_fp + splitRoundPoint.fraction_
					);
			case PointRoundingDirection::None:
				return splitRoundPoint.fraction_;
			default:
				throw LogicError{"Unknown rounding direction supplied for a floating-point."};
			}
		}()
	};
	const auto shiftedRoundedFraction{splitRoundPoint.whole_ + roundedFractionPoint};
	const auto splitRoundedResult{shiftedRoundedFraction.decompose(digitsLeft)};
	const auto roundedNumber{splitNumber.whole_ + (splitRoundedResult.whole_ / shiftOver)};
	return roundedNumber;
}

MajorUnit round(
	const MajorUnit& amount,
	const uint8_t& minorDigits,
	const uint8_t& digitsLeft,
	const function<
							   // roundingDigit,      wholeNumber, isPositive)
		PointRoundingDirection(const MajorUnit&, const MajorUnit&, const bool)
	>& selectDirection
)
{
	if (digitsLeft >= minorDigits)
		throw InvalidArgument{
			"Cannot round at minor digit position " + to_string(digitsLeft)
				+ " when there are only " + to_string(minorDigits) + " minor digits in total."
		};
	const auto minorMask{static_cast<MajorUnit>(pow(10, minorDigits))};
	const auto major{amount / minorMask};
	const auto minor{amount % minorMask};
	const auto roundingAmount{static_cast<MajorUnit>(pow(10, minorDigits - digitsLeft - 1))};
	const auto roundingMask{static_cast<MajorUnit>(pow(10, minorDigits - digitsLeft))};
	const auto roundingPointValue{minor % roundingMask};
	const bool isPositive{amount > 0};
	const auto roundingPointDigit{abs(roundingPointValue / roundingAmount)};
	const auto direction{selectDirection(roundingPointDigit, major, isPositive)};
	const auto unmaskedMajor{major * minorMask};
	const auto roundMinorToNextPoint{
		[&] () -> MajorUnit
		{
			switch (direction)
			{
			case PointRoundingDirection::Up:
				return (isPositive ? (minorMask - minor) : -minor) % roundingMask;
			case PointRoundingDirection::Down:
				return (isPositive ? -minor : (-minorMask - minor)) % roundingMask;
			case PointRoundingDirection::None:
				return 0;
			default:
				throw LogicError{"Unknown rounding direction supplied for a major unit."};
			}
		}()
	};
	const auto roundedRaw{unmaskedMajor + (minor + roundMinorToNextPoint)};
	const auto rounded{roundedRaw / roundingMask * roundingMask};
	return rounded;
}

}

FloatingPoint pecunia::rounders::reals::up(const FloatingPoint& number, const uint8_t& digitsLeft)
{
	return round(
		number,
		digitsLeft,
		[] (const FloatingPoint& roundingDigit, const FloatingPoint&, const bool)
		{
			if (roundingDigit.isEqual(0))
				return PointRoundingDirection::None;
			return PointRoundingDirection::Up;
		}
	);
}

FloatingPoint pecunia::rounders::reals::down(
	const FloatingPoint& number,
	const uint8_t& digitsLeft
)
{
	return round(
		number,
		digitsLeft,
		[] (const FloatingPoint& roundingDigit, const FloatingPoint&, const bool)
		{
			if (roundingDigit.isEqual(0))
				return PointRoundingDirection::None;
			return PointRoundingDirection::Down;
		}
	);
}

FloatingPoint pecunia::rounders::reals::even(
	const FloatingPoint& number,
	const uint8_t& digitsLeft
)
{
	return round(
		number,
		digitsLeft,
		[] (const FloatingPoint& roundingDigit, const FloatingPoint& whole, const bool isPositive)
		{
			if (roundingDigit.isEqual(0))
				return PointRoundingDirection::None;

			if (whole.isEqual(0))
			{
				if (isPositive)
				{
					if (roundingDigit > 0.5)
						return PointRoundingDirection::Up;
					return PointRoundingDirection::Down;
				}

				if (roundingDigit > 0.5)
					return PointRoundingDirection::Down;
				return PointRoundingDirection::Up;
			}
			const auto next{whole > 0 ? whole + 1 : whole};

			if (next.modulo(2).isEqual(0) && (roundingDigit > 0.5 || roundingDigit.isEqual(0.5)))
				return PointRoundingDirection::Up;
			return PointRoundingDirection::Down;
		}
	);
}

FloatingPoint pecunia::rounders::reals::odd(
	const FloatingPoint& number,
	const uint8_t& digitsLeft
)
{
	return round(
		number,
		digitsLeft,
		[] (const FloatingPoint& roundingDigit, const FloatingPoint& whole, const bool isPositive)
		{
			if (roundingDigit.isEqual(0))
				return PointRoundingDirection::None;

			if (whole.isEqual(0))
			{
				if (isPositive)
				{
					if (roundingDigit > 0.5 || roundingDigit.isEqual(0.5))
						return PointRoundingDirection::Up;
					return PointRoundingDirection::Down;
				}

				if (roundingDigit < 0.5)
					return PointRoundingDirection::Up;
				return PointRoundingDirection::Down;
			}

			if (isPositive && roundingDigit > 0.5)
				return PointRoundingDirection::Up;

			if ( ! isPositive && (roundingDigit < 0.5 || roundingDigit.isEqual(0.5)))
				return PointRoundingDirection::Up;
			return PointRoundingDirection::Down;
		}
	);
}

FloatingPoint pecunia::rounders::reals::none(const FloatingPoint& number, const uint8_t&)
{
	return number;
}

MajorUnit pecunia::rounders::currency::up(
	const MajorUnit& amount,
	const uint8_t& minorDigits,
	const uint8_t& digitsLeft
)
{
	return round(
		amount,
		minorDigits,
		digitsLeft,
		[] (const MajorUnit& roundingDigit, const MajorUnit&, const bool)
		{
			if (roundingDigit == 0)
				return PointRoundingDirection::None;
			return PointRoundingDirection::Up;
		}
	);
}

MajorUnit pecunia::rounders::currency::down(
	const MajorUnit& amount,
	const uint8_t& minorDigits,
	const uint8_t& digitsLeft
)
{
	return round(
		amount,
		minorDigits,
		digitsLeft,
		[] (const MajorUnit& roundingDigit, const MajorUnit&, const bool)
		{
			if (roundingDigit == 0)
				return PointRoundingDirection::None;
			return PointRoundingDirection::Down;
		}
	);
}

MajorUnit pecunia::rounders::currency::even(
	const MajorUnit& amount,
	const uint8_t& minorDigits,
	const uint8_t& digitsLeft
)
{
	return round(
		amount,
		minorDigits,
		digitsLeft,
		[] (const MajorUnit& roundingDigit, const MajorUnit& whole, const bool isPositive)
		{
			if (roundingDigit == 0)
				return PointRoundingDirection::None;

			if (whole == 0)
			{
				if (isPositive)
				{
					if (roundingDigit > 5)
						return PointRoundingDirection::Up;
					return PointRoundingDirection::Down;
				}

				if (roundingDigit > 5)
					return PointRoundingDirection::Down;
				return PointRoundingDirection::Up;
			}
			const auto next{whole > 0 ? whole + 1 : whole};

			if (next % 2 == 0 && roundingDigit >= 5)
				return PointRoundingDirection::Up;
			return PointRoundingDirection::Down;
		}
	);
}

MajorUnit pecunia::rounders::currency::none(const MajorUnit& amount, const uint8_t&, const uint8_t&)
{
	return amount;
}
