/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Splitter.h"
using namespace std::string_literals;
using std::string;
using std::to_string;
using std::stoll;
using std::stoull;

#include <algorithm>
using std::find_if;
using std::min;
#include <cctype>
using std::isalpha;
using std::isdigit;
#include <cmath>
using std::pow;
#include <cstddef>
using std::size_t;
#include <cstdint>
using std::int64_t;
using std::uint64_t;
using std::uint8_t;
#include <exception>
using std::exception;
#include <limits>
using std::numeric_limits;
#include <locale>
using std::locale;
using std::moneypunct;
using std::use_facet;
#include <optional>
using std::nullopt;
using std::optional;
#include <stdexcept>
using std::invalid_argument;
using std::out_of_range;
#include <utility>
using std::make_pair;
using std::pair;

#include <pecunia/Codes.h>
using pecunia::currency::Currency;
using pecunia::currency::isoCurrencySize;
using pecunia::currency::minorUnitDigits;
using pecunia::currency::toIso4217Code;
using pecunia::currency::toStdString;
#include <pecunia/Errors.h>
using pecunia::InvalidArgument;
using pecunia::LogicError;
using pecunia::OutOfRange;
using pecunia::OverflowError;
using pecunia::RuntimeError;
#include <pecunia/Types.h>
using pecunia::MajorUnit;
using pecunia::MinorUnit;
using pecunia::MoneyParts;
using pecunia::subMinorDigits;


namespace
{

inline constexpr auto separatorSize{1};

void validateAllCharacters(const string& amount, const char majorMinorSeparator)
{
	if (amount.length() < isoCurrencySize)
		throw InvalidArgument{
			"The amount, " + amount + ", does not contain enough characters to form a valid value."
		};
	const auto invalidChar{
		find_if(
			amount.cbegin(),
			amount.cend(),
			[&majorMinorSeparator] (const auto testChar)
			{
				return ! isalpha(testChar)
					&&  ! isdigit(testChar)
					&& testChar != majorMinorSeparator
					&& testChar != '-';
			}
		)
	};

	if (invalidChar != amount.cend())
		throw InvalidArgument{
			"The amount, " + amount + ", contains an invalid character, '" + *invalidChar + "'."
		};
}

pair<Currency, size_t> extractCurrency(const string& amount)
{
	size_t startingCurrencyIndex{0}, endingCurrencyIndex{0};
	bool foundCurrency{false}, wasShiftedOnce{false};

	for (
		auto testEndingIndex{amount.length() - 1} ;
		testEndingIndex > 0 ;
		testEndingIndex > isoCurrencySize
			? testEndingIndex -= isoCurrencySize
			: testEndingIndex = 0
	)
	{
		bool isTestRangeAllLetters{true};

		for (
			auto testCurrencyIndex{testEndingIndex - isoCurrencySize + 1} ;
			testCurrencyIndex <= testEndingIndex ;
			++testCurrencyIndex
		)
			if (isalpha(amount[testCurrencyIndex]) == 0)
			{
				isTestRangeAllLetters = false;
				break;
			}

		if (isTestRangeAllLetters)
		{
			if ( ! foundCurrency)
			{
				endingCurrencyIndex = testEndingIndex + 1;
				startingCurrencyIndex = endingCurrencyIndex - isoCurrencySize;
				foundCurrency = true;
			}
			else
				throw InvalidArgument{
					"The the text of the supplied amount, "s + amount + ", contains more than one "s
						+ "currency code."s
				};
		}

		// Shifting over to ensure not going out of bounds by always chunking per min currency size.
		if ( ! wasShiftedOnce)
		{
			const auto needToShift{amount.size() % isoCurrencySize};
			testEndingIndex -= needToShift;
			wasShiftedOnce = true;
		}
	}

	if ( ! foundCurrency)
		throw InvalidArgument{
			"The the text of the supplied amount, "s + amount + ", does not contain any currency "s
				+ "code."s
		};
	const auto currency{
		amount.substr(startingCurrencyIndex, endingCurrencyIndex - startingCurrencyIndex)
	};

	try
	{
		return make_pair(toIso4217Code(currency), endingCurrencyIndex);
	}
	catch (const exception& error)
	{
		throw InvalidArgument{
			"The the text of the supplied amount, "s + amount + ", couldn't convert the detected "s
				+ "currency, "s + currency + ", into a valid currency code."s,
			error
		};
	}
}

optional<size_t> findMajorMinorSeparatorIndex(const string& amount, const char majorMinorSeparator)
{
	size_t separatorIndex{};
	bool wasSeparatorFound{false};

	for (size_t index{0} ; index < amount.size() ; ++index)
		if (amount[index] == majorMinorSeparator)
		{
			if (wasSeparatorFound)
				throw InvalidArgument{
					"The the text of the supplied amount, "s + amount + ", contains more than one "s
						+ "major/minor currency separator."s
				};
			wasSeparatorFound = true;
			separatorIndex = index;
		}

	if ( ! wasSeparatorFound)
		return nullopt;
	return separatorIndex;
}

MajorUnit extractMajor(
	const string& amount,
	const size_t majorMinorSeparatorIndex,
	const size_t currencyEndingIndex
)
try
{
	const int64_t majorRaw{
		stoll(
			currencyEndingIndex < majorMinorSeparatorIndex
				? amount.substr(
					currencyEndingIndex,
					majorMinorSeparatorIndex - currencyEndingIndex
				)
				: amount.substr(0, majorMinorSeparatorIndex)
		)
	};

	if (majorRaw > numeric_limits<MajorUnit>::max())
		throw OverflowError{
			"The the text of the supplied amount, "s + amount + ", has a major currency value too "s
				+ "large to fit within the type."s
		};
	return static_cast<MajorUnit>(majorRaw);
}
catch (const out_of_range& error)
{
	throw OutOfRange{
		"The the text of the supplied amount, "s + amount + ", does not fit within a 64-bit"s
			+ "integer."s,
		error
	};
}

optional<uint64_t> tryUint64(const string& str)
try
{
	const size_t expectedDigits{str.length()};
	size_t processed;
	const auto converted{stoull(str, &processed)};

	if (processed != expectedDigits)
		return nullopt;
	return converted;
}
catch (const exception&)
{
	return nullopt;
}

uint64_t extractMinorRaw(
	const string& amount,
	const uint8_t minorUnitDigits,
	const size_t majorMinorSeparatorIndex,
	const size_t currencyEndingIndex
)
{
	if (minorUnitDigits == 0)
		return 0ull;

	if (currencyEndingIndex < majorMinorSeparatorIndex)
	{
		const auto digitsAfterSeparator{amount.size() - majorMinorSeparatorIndex - separatorSize};
		auto extracted{
			// Convert Extract Form: CurrencyMajorSeparatorMinorSubMinor.
			tryUint64(
				amount.substr(
					majorMinorSeparatorIndex + separatorSize,
					digitsAfterSeparator - subMinorDigits
				)
			)
		};

		if (digitsAfterSeparator > static_cast<size_t>(minorUnitDigits + subMinorDigits))
			throw OverflowError{
				"The the text of the supplied amount, "s + amount + ", contains too many digits "s
				+ "after the major/minor currency separator."
			};

		if (extracted.has_value())
			return *extracted;
		throw InvalidArgument{
			"The the text of the supplied amount, "s + amount + ", does not contain a minor "s
				+ "currency value."
		};
	}
	const auto digitsAfterSeparator{
		currencyEndingIndex - majorMinorSeparatorIndex - isoCurrencySize - separatorSize
	};
	auto extracted{
		// Convert Extract Form: MajorSeparatorMinorSubMinorCurrency.
		tryUint64(
			amount.substr(
				majorMinorSeparatorIndex + separatorSize,
				digitsAfterSeparator - subMinorDigits
			)
		)
	};

	if (digitsAfterSeparator > static_cast<size_t>(minorUnitDigits + subMinorDigits))
		throw OverflowError{
			"The the text of the supplied amount, "s + amount + ", contains too many digits "s
				+ "after the major/minor currency separator."
		};

	if (extracted.has_value())
		return *extracted;
	throw InvalidArgument{
		"The the text of the supplied amount, "s + amount + ", does not contain a minor currency "s
			+ "value."
	};
}

MinorUnit extractMinor(
	const string& amount,
	const uint8_t minorUnitDigits,
	const size_t majorMinorSeparatorIndex,
	const size_t currencyEndingIndex
)
{
	const auto minorRaw{
		extractMinorRaw(amount, minorUnitDigits, majorMinorSeparatorIndex, currencyEndingIndex)
	};

	if (minorRaw > numeric_limits<MinorUnit>::max())
		throw OverflowError{
			"The the text of the supplied amount, "s + amount + ", has a minor currency value too "s
				+ "large to fit within the type."s
		};
	return static_cast<MinorUnit>(minorRaw);
}

uint64_t extractSubMinorRaw(
	const string& amount,
	const uint8_t minorUnitDigits,
	const size_t majorMinorSeparatorIndex,
	const size_t currencyEndingIndex
)
{
	const auto subMinorUnitIndex{majorMinorSeparatorIndex + minorUnitDigits + separatorSize};

	if (currencyEndingIndex < majorMinorSeparatorIndex)
	{
		// Convert Extract Form: CurrencyMajorSeparatorMinorSubMinor.
		auto extracted{tryUint64(amount.substr(subMinorUnitIndex, subMinorDigits))};

		if (extracted.has_value())
			return *extracted;
		throw InvalidArgument{
			"The the text of the supplied amount, "s + amount + ", does not contain a valid "s
				+ "sub-minor currency value."
		};
	}
	// Convert Extract Form: MajorSeparatorMinorSubMinorCurrency.
	auto extracted{tryUint64(amount.substr(subMinorUnitIndex, subMinorDigits))};

	if (extracted.has_value())
		return *extracted;
	throw InvalidArgument{
		"The the text of the supplied amount, "s + amount + ", does not contain a full valid "s
			+ "sub-minor currency value."
	};
}

MinorUnit extractSubMinor(
	const string& amount,
	const uint8_t minorUnitDigits,
	const size_t majorMinorSeparatorIndex,
	const size_t currencyEndingIndex
)
{
	const auto subMinorRaw{
		extractSubMinorRaw(amount, minorUnitDigits, majorMinorSeparatorIndex, currencyEndingIndex)
	};

	if (subMinorRaw > numeric_limits<MinorUnit>::max())
		throw OverflowError{
			"The the text of the supplied amount, "s + amount + ", has a sub-minor currency value "s
				+ "too large to fit within the type."s
		};
	return static_cast<MinorUnit>(subMinorRaw);
}

string normaliseZeros(
	const string& original,
	const size_t expectedDigits,
	const char majorMinorSeparator,
	const optional<size_t>& separatorIndex,
	size_t& currencyEndingIndex
)
{
	const bool currencyAtEnd{currencyEndingIndex == original.size()};
	const size_t majorMinorSeparatorIndex{
		separatorIndex.value_or(
			currencyAtEnd
				? original.size() - isoCurrencySize - separatorSize
				: original.size() - separatorSize
		)
	};
	// Take all major, major/minor separator, and possibly currency characters.
	string normalised{original.substr(0, majorMinorSeparatorIndex + separatorSize)};

	if ( ! separatorIndex.has_value())
		normalised += majorMinorSeparator;
	const auto digitsAfterSeparator{
		original.length() - (currencyAtEnd ? isoCurrencySize : 0) - majorMinorSeparatorIndex
			- separatorSize
	};
	// Add remaining digits.
	normalised += original.substr(majorMinorSeparatorIndex + separatorSize, digitsAfterSeparator);
	const auto zerosToAdd{
		expectedDigits > digitsAfterSeparator
			? expectedDigits - digitsAfterSeparator
			: 0
	};

	// Add all zeros not entered.
	if (zerosToAdd > 0)
		normalised += string(zerosToAdd, '0');

	// Add the currency if it appeared at the end.
	if (currencyAtEnd)
	{
		normalised += original.substr(currencyEndingIndex - isoCurrencySize, isoCurrencySize);
		currencyEndingIndex += zerosToAdd;
	}
	return normalised;
}

}

MoneyParts pecunia::split(const string& amount)
{
	const locale loc{};
	const auto majorMinorSeparator{use_facet<moneypunct<char>>(loc).decimal_point()};
	return split(amount, majorMinorSeparator);
}

MoneyParts pecunia::split(const string& amount, const char majorMinorSeparator)
{
	validateAllCharacters(amount, majorMinorSeparator);
	auto [currency, currencyEndingIndex]{extractCurrency(amount)};
	auto majorMinorSeparatorIndex{findMajorMinorSeparatorIndex(amount, majorMinorSeparator)};
	const auto minorDigits{minorUnitDigits(currency)};
	const auto normalised{
		normaliseZeros(
			amount,
			minorDigits + subMinorDigits,
			majorMinorSeparator,
			majorMinorSeparatorIndex,
			currencyEndingIndex
		)
	};

	if ( ! majorMinorSeparatorIndex.has_value())
		majorMinorSeparatorIndex = findMajorMinorSeparatorIndex(normalised, majorMinorSeparator);
	const auto major{extractMajor(normalised, *majorMinorSeparatorIndex, currencyEndingIndex)};
	const auto minorOnly{
		extractMinor(normalised, minorDigits, *majorMinorSeparatorIndex, currencyEndingIndex)
	};
	const auto subMinor{
		extractSubMinor(normalised, minorDigits, *majorMinorSeparatorIndex, currencyEndingIndex)
	};
	return {currency, major, minorDigits == 0 ? nullopt : optional{minorOnly}, subMinor};
}
