/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_SPLITTER_H
#define PECUNIA_SPLITTER_H

#include <string>

#include <pecunia/pecunia_export.h>


namespace pecunia
{

class MoneyParts;

/**
 * @brief Splits a string into the various parts of a monetary value. Performance wise it will
 * favour strings with a trailing currency. Uses the systems locale to determine the major and
 * minor separator. Valid forms are:
 *	- CurrencyMajorSeparatorMinorSub-minor
 * 	- CurrencyMajorSeparatorMinor
 * 	- CurrencyMajorSeparator
 * 	- CurrencyMajor
 *	- MajorSeparatorMinorSub-minorCurrency
 * 	- MajorSeparatorMinorCurrency
 * 	- MajorSeparatorCurrency
 * 	- MajorCurrency
 *
 * @exception InvalidArgument When the string does not contain a valid currency.
 * @exception InvalidArgument When the string contains an invalid major/minor separator.
 * @exception OverflowError  When the string contains a major value too large to store.
 * @exception OutOfRange When the string contains a major value too large to store temporally.
 * @exception InvalidArgument When the string contains an invalid major value.
 * @exception InvalidArgument When the string contains an invalid minor value.
 * @exception InvalidArgument  When the string does not contain a minor value and should.
 * @exception OverflowError When the string contains a minor value too large to store.
 * @exception OutOfRange When the string contains a minor value too large to store temporally.
 * @exception InvalidArgument When the string contains an invalid minor value.
 * @exception InvalidArgument  When the string does not contain a valid sub-minor value.
 * @exception OverflowError When the string contains a sub-minor value too large to store.
 *
 * @param amount The sequence of characters to split as a monetary value.
 *
 * @return Ever part of the monetary value split apart.
 */
[[nodiscard]]
PECUNIA_EXPORT MoneyParts split(const std::string& amount);
/**
 * @brief Splits a string into the various parts of a monetary value. Performance wise it will
 * favour strings with a trailing currency. Valid forms are:
 *	- CurrencyMajorSeparatorMinorSub-minor
 * 	- CurrencyMajorSeparatorMinor
 * 	- CurrencyMajorSeparator
 * 	- CurrencyMajor
 *	- MajorSeparatorMinorSub-minorCurrency
 * 	- MajorSeparatorMinorCurrency
 * 	- MajorSeparatorCurrency
 * 	- MajorCurrency
 *
 * @exception InvalidArgument When the string does not contain a valid currency.
 * @exception InvalidArgument When the string contains an invalid major/minor separator.
 * @exception OverflowError  When the string contains a major value too large to store.
 * @exception OutOfRange When the string contains a major value too large to store temporally.
 * @exception InvalidArgument When the string contains an invalid major value.
 * @exception InvalidArgument When the string contains an invalid minor value.
 * @exception InvalidArgument  When the string does not contain a minor value and should.
 * @exception OverflowError When the string contains a minor value too large to store.
 * @exception OutOfRange When the string contains a minor value too large to store temporally.
 * @exception InvalidArgument When the string contains an invalid minor value.
 * @exception InvalidArgument  When the string does not contain a valid sub-minor value.
 * @exception OverflowError When the string contains a sub-minor value too large to store.
 *
 * @param amount The sequence of characters to split as a monetary value.
 * @param majorMinorSeparator The character expected as the separator between a major and
 * minor values.
 *
 * @return Ever part of the monetary value split apart.
 */
[[nodiscard]]
PECUNIA_EXPORT MoneyParts split(const std::string& amount, const char majorMinorSeparator);

}

#endif
