/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Types.h"
using pecunia::ConversionFunction;
using pecunia::MoneyParts;
using pecunia::defaultConverter;
using std::optional;
using std::nullopt;

#include <cmath>
using std::pow;
#include <type_traits>
using std::is_floating_point;
using std::is_integral;
using std::is_signed;
using std::is_unsigned;

#include <internal/Adjustments.h>
using pecunia::currency::internal::minorUnitSubMinorFactor;
#include <pecunia/Codes.h>
using pecunia::currency::Currency;
using pecunia::currency::minorUnitDigits;
using pecunia::currency::toStdString;
#include <pecunia/Errors.h>
using pecunia::LogicError;
using pecunia::RuntimeError;
#include <pecunia/FloatingPoint.h>
using namespace pecunia::literals;
using pecunia::FloatingPoint;
#include <pecunia/Sizes.hpp>
using pecunia::FloatingPointBase;
using pecunia::MajorUnit;
using pecunia::MinorUnit;
using pecunia::subMinorDigits;


namespace
{

MinorUnit toMinor(
	const optional<MinorUnit> minorOnly,
	const MinorUnit subMinor
)
{
	if ( ! minorOnly.has_value())
		return subMinor;
	const auto subMinorFactor{static_cast<MinorUnit>(pow(10, subMinorDigits))};
	auto minor{static_cast<MinorUnit>((*minorOnly * subMinorFactor) + subMinor)};
	return minor;
}

MajorUnit toFull(
	const MajorUnit major,
	const optional<MinorUnit> minorOnly,
	const MinorUnit subMinor,
	const Currency code
)
{
	const auto majorAdjustment{static_cast<MajorUnit>(minorUnitSubMinorFactor(code))};
	const auto minor{toMinor(minorOnly, subMinor)};
	auto full{
		static_cast<MajorUnit>(
			major >= 0
				? (major * majorAdjustment) + minor
				: (major * majorAdjustment) - minor
		)
	};
	return full;
}

}

static_assert(is_unsigned<MinorUnit>::value, "The MinorUnit must be unsigned.");
static_assert(is_integral<MinorUnit>::value, "The MinorUnit must be an integer.");
static_assert(
	is_floating_point<FloatingPointBase>::value,
	"The FloatingPointBase is not a floating-point type."
);
static_assert(is_signed<MajorUnit>::value, "The MajorUnit must be signed.");
static_assert(is_integral<MajorUnit>::value, "The MajorUnit must be an integral.");

MoneyParts::MoneyParts() :
	MoneyParts{0, 0, 0u, nullopt, 0u, Currency::XXX}
{}

MoneyParts::MoneyParts(
	const MajorUnit full,
	const MajorUnit major,
	const MinorUnit minor,
	const optional<MinorUnit> minorOnly,
	const MinorUnit subMinor,
	const Currency code
) :
	full_{full},
	major_{major},
	minorAndSub_{minor},
	minorOnly_{minorOnly},
	subMinor_{subMinor},
	code_{code}
{
	if (this->minorOnly_.has_value() && minorUnitDigits(this->code_) == 0)
		throw InvalidArgument{
			"The currency '" + toStdString(this->code_) + "' does not support minor values."
		};
}

MoneyParts::MoneyParts(
	const Currency code,
	const MajorUnit major,
	const optional<MinorUnit> minorOnly,
	const MinorUnit subMinor
) :
	MoneyParts{
		toFull(major, minorOnly, subMinor, code),
		major,
		toMinor(minorOnly, subMinor),
		minorOnly,
		subMinor,
		code
	}
{}

const MajorUnit& MoneyParts::full() const noexcept
{
	return this->full_;
}

const MajorUnit& MoneyParts::major() const noexcept
{
	return this->major_;
}

const MinorUnit& MoneyParts::minorAndSub() const noexcept
{
	return this->minorAndSub_;
}

const optional<MinorUnit>& MoneyParts::minorOnly() const noexcept
{
	return this->minorOnly_;
}

const MinorUnit& MoneyParts::subMinor() const noexcept
{
	return this->subMinor_;
}

const Currency& MoneyParts::code() const noexcept
{
	return this->code_;
}

FloatingPoint pecunia::defaultConverter(const Currency& from, const Currency& to)
{
	if (from == to)
		return 1.0_fp;
	throw LogicError{
		"Default converter only handles conversion when the currencies match."
	};
}

bool pecunia::operator==(const MoneyParts& lhs, const MoneyParts& rhs)
{
	return lhs.full() == rhs.full()
		&& lhs.major() == rhs.major()
		&& lhs.minorAndSub() == rhs.minorAndSub()
		&& lhs.minorOnly() == rhs.minorOnly()
		&& lhs.subMinor() == rhs.subMinor()
		&& lhs.code() == rhs.code();
}

ConversionFunction pecunia::converter{&defaultConverter};
