/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_TYPES_H
#define PECUNIA_TYPES_H

#include <cstdint>
#include <functional>
#include <optional>

#ifdef __WIN32
#	pragma STDC FENV_ACCESS ON
#endif

#include <pecunia/pecunia_export.h>
#include <pecunia/Sizes.hpp>


namespace pecunia
{

class FloatingPoint;

}
namespace pecunia::currency
{

enum class Currency : std::uint16_t;
class Money;

}
namespace pecunia
{

/**
 * @brief When working only with minor units, this represents the positivity or negativity of a
 * minor value.
 */
enum class PECUNIA_EXPORT MinorSign
{
	/// The minor value is negative.
	Negative,
	/// The minor value is neither positive nor negative, i.e. zero.
	Neither,
	/// The minor value is positive.
	Positive,
};
/**
 * @brief A representation of a function that performs a rounding operation upon a real number. This
 * function must take a number, the number of digits after the whole number to round, and giving
 * back the rounded number.
 *
 * @exception pecunia::RuntimeError When the rounding is not possible.
 */
using FloatingPointRounderFunction =
		std::function<FloatingPoint(const FloatingPoint&, const std::uint8_t&)>;
/**
 * @brief A representation of a function that performs a rounding operation upon a money storage
 * type. This function must take first the major and minor amount value stored together, then the
 * number of digits in the minor unit, finally the number of digits after the whole number to round,
 * and giving back the rounded number with the major and minor amount stored together.
 *
 * @exception pecunia::RuntimeError When the rounding is not possible.
 */
using RounderFunction =
		std::function<MajorUnit(const MajorUnit&, const std::uint8_t&, const std::uint8_t)>;
/**
 * @brief A conversion function between two currencies where the first argument is the currency to
 * convert from, the second argument is the currency to convert to, and giving back the conversion
 * factor to get from the first currency to the second currency.
 *
 * @exception pecunia::RuntimeError When either the conversion from or to a currency is not supported.
 */
using ConversionFunction =
		std::function<FloatingPoint(const currency::Currency&, const currency::Currency&)>;

/**
 * @brief A representation of all parts of a monetary value.
 */
class MoneyParts
{
	// Justification: Used for performance reasons.
	friend class pecunia::currency::Money;

	MoneyParts(
		const MajorUnit full,
		const MajorUnit major,
		const MinorUnit minor,
		const std::optional<MinorUnit> minorOnly,
		const MinorUnit subMinor,
		const pecunia::currency::Currency code
	);

	MajorUnit full_;
	MajorUnit major_;
	MinorUnit minorAndSub_;
	std::optional<MinorUnit> minorOnly_;
	MinorUnit subMinor_;
	pecunia::currency::Currency code_;

public:
	PECUNIA_EXPORT MoneyParts();
	/**
	 * @brief Component constructor that fully initialises the object.
	 *
	 * @exception InvalidArgument When the monetary value does not support minor units and minor
	 * units are supplied.
	 *
	 * @param code The currency of the monetary value.
	 * @param major The major monetary value.
	 * @param minorOnly The minor monetary value, if the currency supports it.
	 * @param subMinor The sub-minor monetary value.
	 */
	PECUNIA_EXPORT MoneyParts(
		const pecunia::currency::Currency code,
		const MajorUnit major,
		const std::optional<MinorUnit> minorOnly,
		const MinorUnit subMinor = 0
	);
	PECUNIA_EXPORT MoneyParts(const MoneyParts&) = default;
	PECUNIA_EXPORT MoneyParts(MoneyParts&&) = default;
	PECUNIA_EXPORT MoneyParts& operator=(const MoneyParts&) = default;
	PECUNIA_EXPORT MoneyParts& operator=(MoneyParts&&) = default;

	/// An accessor to the full amount monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const MajorUnit& full() const noexcept;
	/// An accessor to the major monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const MajorUnit& major() const noexcept;
	/// An accessor to the minor and sub-minor monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const MinorUnit& minorAndSub() const noexcept;
	/// An accessor to the minor monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const std::optional<MinorUnit>& minorOnly() const noexcept;
	/// An accessor to the sub-minor monetary value.
	[[nodiscard]]
	PECUNIA_EXPORT const MinorUnit& subMinor() const noexcept;
	/// An accessor to the ISO currency code of the monetary amount.
	[[nodiscard]]
	PECUNIA_EXPORT const pecunia::currency::Currency& code() const noexcept;
};

/**
 * @brief Provides a 1:1 conversion for a currency whose currency::Currency match.
 *
 * @param[in] from The currency code whose conversion value is desired.
 * @param[in] to The currency code the conversion value is directed towards.
 *
 * @throw LogicError When either the from and to currency currency::Currency do not match.
 *
 * @return The conversion value for the supplied currency when multiplied from to the other.
 */
[[nodiscard]]
PECUNIA_EXPORT FloatingPoint defaultConverter( // TODO: rename convert or something
	const currency::Currency& from,
	const currency::Currency& to
);
/**
 * @brief General equality operator.
 *
 * @param lhs The object that is expected to be compared.
 * @param rhs The object that is to compare for equality.
 *
 * @return True when all member variables are the same, false otherwise.
 */
[[nodiscard]]
PECUNIA_EXPORT bool operator==(const MoneyParts& lhs, const MoneyParts& rhs);

/// The holder for all conversion operations for any money unit.
PECUNIA_EXPORT extern ConversionFunction converter;

}

#endif
