/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Adjustments.h"
using pecunia::FloatingPointRounderFunction;
using pecunia::MajorUnit;
using pecunia::MinorUnit;
using pecunia::converter;
using pecunia::currency::internal::minorUnitSubMinorFactor;
using pecunia::subMinorDigits;
using std::function;
using std::int32_t;
using std::uint8_t;

#include <cassert>
// using assert;
#include <cmath>
using std::abs;

#include <internal/Sizes.hpp>
using pecunia::internal::extraFloatingPointPrecision;
#include <pecunia/Codes.h>
using pecunia::currency::Currency;
#include <pecunia/FloatingPoint.h>
using namespace pecunia::literals;
using pecunia::FloatingPoint;
#include <pecunia/Money.h>
using pecunia::currency::Money;
using pecunia::currency::toStdString;
#include <pecunia/Types.h>
using pecunia::subMinorDigits;


namespace
{

MinorUnit calculateFloatingPointMinorAndSubMinor(
	const FloatingPoint& value,
	const Currency& code,
	const uint8_t subMinorUnitDigits,
	const FloatingPointRounderFunction& rounder
)
{
	assert(rounder != nullptr && "The floating point rounder must be a function.");
	const uint8_t totalDigits{
		static_cast<uint8_t>(
			minorUnitDigits(code) + subMinorUnitDigits + extraFloatingPointPrecision
		)
	};
	const int32_t ratio{
		minorUnitSubMinorFactor(code)
			* static_cast<int32_t>(pow(10, extraFloatingPointPrecision))
	};
	const FloatingPoint rounded{rounder(value, totalDigits)};
	const MajorUnit adjustedValue{static_cast<MajorUnit>((rounded * ratio).underlying())};
	const auto minorAdjusted{abs(adjustedValue) % ratio};
	const auto extraRatio{pow(10.0, extraFloatingPointPrecision)};
	const auto removedExtra{minorAdjusted / static_cast<MinorUnit>(extraRatio)};
	return static_cast<MinorUnit>(removedExtra);
}

}

MinorUnit pecunia::currency::internal::fromMinorMinorOnly(
	const MinorUnit minorAndSub,
	const uint8_t subMinorUnitDigits
)
{
	const auto subMinorAdjustor{static_cast<MinorUnit>(pow(10, subMinorUnitDigits))};
	const auto justMinor{minorAndSub / subMinorAdjustor};
	return static_cast<MinorUnit>(justMinor);
}

MinorUnit pecunia::currency::internal::fromMinorSubMinor(
	const MinorUnit minorAndSub,
	const uint8_t subMinorUnitDigits
)
{
	const auto subMinorAdjustor{static_cast<MinorUnit>(pow(10, subMinorUnitDigits))};
	const auto justSub{minorAndSub % subMinorAdjustor};
	return static_cast<MinorUnit>(justSub);
}

MajorUnit pecunia::currency::internal::fromFloatingPointMajor(
	const FloatingPoint& value,
	const Currency& code,
	const uint8_t subMinorUnitDigits,
	const FloatingPointRounderFunction& rounder
)
{
	assert(rounder != nullptr && "The floating point rounder must be a function.");
	const uint8_t totalDigits{
		static_cast<uint8_t>(
			minorUnitDigits(code) + subMinorUnitDigits + extraFloatingPointPrecision
		)
	};
	const int32_t ratio{minorUnitSubMinorFactor(code)};
	const FloatingPoint rounded{rounder(value, totalDigits)};
	const MajorUnit major{static_cast<MajorUnit>(((rounded * ratio) / ratio).underlying())};
	return major;
}

MinorUnit pecunia::currency::internal::fromFloatingPointMinor(
	const FloatingPoint& value,
	const Currency& code,
	const uint8_t subMinorUnitDigits,
	const FloatingPointRounderFunction& rounder
)
{
	const auto minorAndSub{
		calculateFloatingPointMinorAndSubMinor(value, code, subMinorUnitDigits, rounder)
	};
	return fromMinorMinorOnly(minorAndSub, subMinorUnitDigits);
}

MinorUnit pecunia::currency::internal::fromFloatingPointSubMinor(
	const FloatingPoint& value,
	const Currency& code,
	const uint8_t subMinorUnitDigits,
	const FloatingPointRounderFunction& rounder
)
{
	const auto minorAndSub{
		calculateFloatingPointMinorAndSubMinor(value, code, subMinorUnitDigits, rounder)
	};
	return fromMinorSubMinor(minorAndSub, subMinorUnitDigits);
}

MajorUnit pecunia::currency::internal::normaliseAmount(const Money& from, const Currency& to)
{
	if (from.code() == to)
		return from.amount();
	assert(converter != nullptr && "The currency conversion function is not set.");
	const FloatingPoint conversionFactor{converter(from.code(), to)};

	if (
		conversionFactor.isEqual(
			1.0,
			static_cast<uint8_t>(minorUnitDigits(from.code()) + subMinorDigits)
		)
	)
		return from.amount();
	const auto convertedAmount{FloatingPoint{from.amount()} * conversionFactor};
	return static_cast<MajorUnit>(convertedAmount.underlying());
}

MajorUnit pecunia::currency::internal::fromAmountMajor(
	const MajorUnit& amount,
	const Currency& code
)
{
	return amount / minorUnitSubMinorFactor(code);
}

MinorUnit pecunia::currency::internal::fromAmountMinor(
	const MajorUnit& amount,
	const Currency& code
)
{
	return static_cast<MinorUnit>(abs(amount % minorUnitSubMinorFactor(code)));
}

MinorUnit pecunia::currency::internal::fromAmountMinorOnly(
	const MajorUnit& amount,
	const Currency& code
)
{
	return fromMinorMinorOnly(fromAmountMinor(amount, code), subMinorDigits);
}

MinorUnit pecunia::currency::internal::fromAmountSubMinor(
	const MajorUnit& amount,
	const Currency& code
)
{
	return fromMinorSubMinor(fromAmountMinor(amount, code), subMinorDigits);
}

MinorUnit pecunia::currency::internal::minorUnitSubMinorFactor(const Currency& code) noexcept
{
	return static_cast<MinorUnit>(
		10.0_fp .power(minorUnitDigits(code) + subMinorDigits).underlying()
	);
}
