/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Verification.h"
using pecunia::FloatingPointBase;
using pecunia::MajorUnit;
using pecunia::MinorUnit;
using pecunia::internal::VerifiedValue;
using std::int32_t;
using std::int64_t;
using std::uint8_t;

#include <pecunia/FloatingPoint.h>
using namespace pecunia::literals;
using pecunia::FloatingPoint;


VerifiedValue pecunia::internal::verifyAdditionFits(
	const MajorUnit& lhs,
	const MajorUnit& rhs,
	const MajorUnit& maximum,
	const MajorUnit& minimum
)
{
	if (lhs > 0 && rhs > (maximum - lhs))
		return VerifiedValue::Overflows;

	if (lhs < 0 && rhs < (minimum - lhs))
		return VerifiedValue::Overflows;
	return VerifiedValue::Fits;
}

VerifiedValue pecunia::internal::verifySubtractionFits(
	const MajorUnit& lhs,
	const MajorUnit& rhs,
	const MajorUnit& maximum,
	const MajorUnit& minimum
)
{
	// N.B. cannot negate when at the minimum.
	if (rhs == minimum)
		return VerifiedValue::Overflows;
	return verifyAdditionFits(lhs, -rhs, maximum, minimum);
}

VerifiedValue pecunia::internal::verifyMultiplicationFits(
	const MajorUnit& lhs,
	const FloatingPoint& rhs,
	const MajorUnit& maximum,
	const MajorUnit& minimum,
	const MinorUnit& minorMaximum,
	const uint8_t minorPrecision
)
{
	if ((lhs == 0) || rhs.isEqual(0, minorPrecision))
		return VerifiedValue::Fits;

	if (
		(lhs == -1 && rhs.isEqual(minimum, minorPrecision))
		|| (rhs.isEqual(-1, minorPrecision) && lhs == minimum)
	)
		return VerifiedValue::Overflows;

	if (rhs.isEqual(-1, minorPrecision))
		return VerifiedValue::Fits;
	const auto fpLhs{static_cast<FloatingPointBase>(lhs)};
	const auto fpMaximum{static_cast<FloatingPointBase>(maximum)};

	if (fpLhs > fpMaximum / rhs)
		return VerifiedValue::Overflows;
	const auto fpMinimum{static_cast<FloatingPointBase>(minimum)};

	if (fpLhs < fpMinimum / rhs)
		return VerifiedValue::Overflows;

	if (rhs < 1 && rhs > -1)
	{
		return verifyDivisionFits(
			lhs,
			rhs.reciprocal(),
			maximum,
			minimum,
			minorMaximum,
			minorPrecision
		);
	}
	return VerifiedValue::Fits;
}

VerifiedValue pecunia::internal::verifyDivisionFits(
	const MajorUnit& dividend,
	const FloatingPoint& divisor,
	const MajorUnit& maximum,
	const MajorUnit& minimum,
	const MinorUnit& minorMaximum,
	const uint8_t minorPrecision
)
{
	if (dividend == 0)
		return VerifiedValue::Fits;

	if (divisor.isEqual(0, minorPrecision))
		return VerifiedValue::Undefined;

	if (divisor < 1 && divisor > -1)
	{
		return verifyMultiplicationFits(
			dividend,
			divisor.reciprocal(),
			maximum,
			minimum,
			minorMaximum,
			minorPrecision
		);
	}

	if (divisor.isEqual(-1, minorPrecision))
	{
		if (dividend < minimum)
			return VerifiedValue::Overflows;
		else
			return VerifiedValue::Fits;
	}
	const FloatingPoint dividendFp{dividend};

	if (dividendFp < divisor)
	{
		const auto resultCheck{dividendFp / divisor};

		if ((resultCheck < 1.0_fp) && (resultCheck > -1.0_fp))
			return VerifiedValue::Underflows;
	}
	return VerifiedValue::Fits;
}

VerifiedValue pecunia::internal::verifyMultiplicationFits(
	const MajorUnit& lhs,
	const MajorUnit& rhs,
	const MajorUnit& maximum,
	const MajorUnit& minimum
)
{
	if (lhs == 0 || rhs == 0)
		return VerifiedValue::Fits;

	if ((lhs == -1 && rhs < minimum) || (rhs == -1 && lhs < minimum))
		return VerifiedValue::Overflows;

	if (rhs == -1)
		return VerifiedValue::Fits;

	if (lhs > maximum / rhs)
		return VerifiedValue::Overflows;

	if (lhs < minimum / rhs)
		return VerifiedValue::Overflows;
	return VerifiedValue::Fits;
}

VerifiedValue pecunia::internal::verifyDivisionFits(
	const MajorUnit& dividend,
	const MajorUnit& divisor,
	const MajorUnit& minimum
)
{
	if (divisor == 0)
		return VerifiedValue::Undefined;

	if (dividend == 0)
		return VerifiedValue::Fits;

	if (divisor == -1)
	{
		if (dividend < minimum)
			return VerifiedValue::Overflows;
		else
			return VerifiedValue::Fits;
	}
	const FloatingPoint dividendFp{dividend};
	const FloatingPoint divisorFp{divisor};

	if (dividendFp < divisorFp)
	{
		const auto resultCheck{dividendFp / divisorFp};

		if ((resultCheck < 1.0_fp) && (resultCheck > -1.0_fp))
			return VerifiedValue::Underflows;
	}
	return VerifiedValue::Fits;
}
